﻿using System ;
using System.Collections.Generic ;
using MCScheduling.MixedCriticality ;

/**
 * Contains the algorithm to solve mixed-criticality scheduling problems that is an adaptation of a clairvoyant
 * non-preemptive scheduling algorithm by Cecilia Ekelin.
 */
namespace MCScheduling.ClairvoyantEDF
{
    /**
     * The mixed-criticality clairvoyant EDF algorithm implementation.
     * <p>
     * This algorithm is based on the algorithm proposed by Cecilia Ekelin, 
     * and also described by the same person in the article entitled
     * as Clairvoyant Non-preemptive EDF Scheduling, which may be
     * found in proceedings of the 18th Euromicro Conference 
     * on Real-Time Systems.
     * </p>
     * 
     * @author Tomas Tunys
     */
    public class ClairvoyantEDF
    {
        /**
         * The comparer used to sort task according to
         * their arrival times.
         * 
         * @author Tomas Tunys
         */
        private class ArrivalComparer : Comparer<Task>
        {
            /**
             * Compares the pair of the specified taks. The smaller
             * is considered the one with the lower arrival time.
             * 
             * @param x The first task to compare.
             * @param y The second task to compare.
             */
            public override int Compare(Task x, Task y)
            {
                return (x.ReleaseTime - y.ReleaseTime) ;
            }
        }

        /**
         * The comparer used to sort task according to
         * their deadlines.
         * 
         * @author Tomas Tunys
         */
        private class DeadlineComparer : Comparer<Task>
        {
            /**
             * Compares the pair of the specified taks. The smaller
             * is considered the one with the lower deadline.
             * 
             * @param x The first task to compare.
             * @param y The second task to compare.
             */
            public override int Compare(Task x, Task y)
            {
                return (x.Deadline - y.Deadline) ;
            }
        }

        /** The queue of tasks sorted according to their arrival times. */
        private PriorityQueue<Task> arrivalQueue = new PriorityQueue<Task>(new ArrivalComparer()) ;

        /** The queue of ready tasks (ones that may be executed) sorted according to their deadlines. */
        private PriorityQueue<Task> readyQueue = new PriorityQueue<Task>(new DeadlineComparer()) ;

        /** The list of tasks sorted according to their maximum start time value. */
        private CriticalList criticalList = new CriticalList() ;

        /**
         * Solves the specified mixed-criticality instance using clairvoyant non-preemptive
         * earliest deadline first algorithm.
         * 
         * @param instance The mixed-criticality instance to solve.
         */
        public void solve(CMixedCriticalityInstance instance)
        {
            foreach (CMixedCriticalityJob job in instance)
            {
                Task task = new Task(job) ;
                arrivalQueue.Add(task) ;
                criticalList.Add(task) ;
            }

            int[] scheduleLength = new int[instance.getMaximumCriticality() + 1] ;

            int processorTime = 0 ;

            while (criticalList.Count != 0)
            {
                if (readyQueue.Count == 0)
                {
                    processorTime = Math.Max(arrivalQueue.GetMinimum().ReleaseTime, processorTime) ;
                }

                Task readyTask = null ;

                while (arrivalQueue.Count != 0 && (readyTask = arrivalQueue.GetMinimum()).ReleaseTime <= processorTime)
                {
                    readyQueue.Add(readyTask) ;
                    arrivalQueue.RemoveMinimum() ;
                }

                Task Ti = readyQueue.GetMinimum() ;

                readyQueue.RemoveMinimum() ;

                int minimumStartTimeAdjustment = Math.Max(scheduleLength[Ti.Criticality], processorTime) ;

                if (Ti.MinimumStartTime < minimumStartTimeAdjustment)
                {
                    Ti.MinimumStartTime = minimumStartTimeAdjustment ;
                }

                Task Tj = criticalList.GetMinimum() ;

                minimumStartTimeAdjustment = Math.Max(scheduleLength[Tj.Criticality], processorTime) ;

                if (Tj.MinimumStartTime < minimumStartTimeAdjustment)
                {
                    Tj.MinimumStartTime = minimumStartTimeAdjustment ;
                }

                int maximumStartTimeBreach = Ti.MinimumStartTime + Ti.ExecutionTime[Math.Min(Ti.Criticality, Tj.Criticality)] ;

                if (maximumStartTimeBreach > Tj.MaximumStartTime && !Object.ReferenceEquals(Ti, Tj) && Tj.MinimumStartTime <= Tj.MaximumStartTime)
                {
                    Ti.MinimumStartTime = Tj.MinimumStartTime + Tj.ExecutionTime[Math.Min(Ti.Criticality, Tj.Criticality)] ;

                    Ti.ReleaseTime = Ti.MinimumStartTime ;

                    arrivalQueue.Add(Ti) ;

                    criticalList.Update(maximumStartTimeBreach, Ti.MaximumStartTime, Ti.Criticality) ;
                }
                else
                {
                    Ti.StartTime = Ti.MinimumStartTime ;

                    for (int criticality = 1 ; criticality <= Ti.Criticality ; criticality++)
                    {
                        scheduleLength[criticality] = Ti.StartTime + Ti.ExecutionTime[criticality] ;
                    }

                    for (int criticality = Ti.Criticality + 1 ; criticality < scheduleLength.Length ; criticality++)
                    {
                        if (scheduleLength[criticality] < scheduleLength[criticality - 1])
                        {
                            scheduleLength[criticality] = scheduleLength[criticality - 1] ;
                        }
                        else
                        {
                            break ;
                        }
                    }

                    processorTime = scheduleLength[1] ;

                    criticalList.Remove(Ti) ;
                }
            }
        }
    }
}
