﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.ClairvoyantEDF
{
    /**
     * A so-called critical list that is used by the clairvoyant EDF
     * algorithm. See ClairvoyantEDF class.
     */
    class CriticalList
    {
        /** The list of tasks of this critical list. */
        private List<Task> taskList = new List<Task>() ;

        /** The number of tasks in the list. */
        public int Count { get { return taskList.Count ; } }

        /**
         * Adds the specified task into the list.
         * 
         * @param task The task to add.
         */
        public void Add(Task task)
        {
            taskList.Add(task) ;
        }

        /**
         * Adds the specified task into the list. Moreover it
         * sets the maximum start times of all the jobs in the
         * list whose maximum start time value is less then 
         * the value of the specified task to maxStartTimeNote.
         * 
         * @param task             The task to add.
         * @param maxStartTimeNote The maximum start time update
         *                         for all tasks whose value of
         *                         this property is less then
         *                         the one of the task to add.
         */
        public void AddWithNote(Task task, int maxStartTimeNote)
        {
            foreach(Task ttask in taskList)
            {
                if (ttask.MaximumStartTime < task.MaximumStartTime)
                {
                    ttask.MaximumStartTime = maxStartTimeNote ;
                }
            }

            task.MaximumStartTime = maxStartTimeNote ;

            Add(task) ;
        }

        /**
         * Updates the maximum start time of each task T whose value of this
         * property is less then maximumStartTimeBreach, such that:
         * 
         * T.MaximumStartTime = min(T.MaximumStartTime,
         * MaximumStartTimeUpdate - T.ExecutionTime[min(T.criticality, criticality)].
         * 
         * @param maximumStartTimeBreach The maximum start time to update.
         * @param maximumStartTimeUpdate The update start time value.
         * @param criticality            The criticality update parameter.
         */
        public void Update(int maximumStartTimeBreach, int maximumStartTimeUpdate, int criticality)
        {
            foreach(Task task in taskList)
            {
                if (task.MaximumStartTime < maximumStartTimeBreach)
                {
                    task.MaximumStartTime = Math.Min(task.MaximumStartTime, maximumStartTimeUpdate - task.ExecutionTime[Math.Min(task.Criticality, criticality)]) ;
                }
            }
        }

        /**
         * Returns the taks with the lowest maximal start time.
         * 
         * @return The taks with the lowest maximal start time.
         * 
         * @throws InvalidOperationException If the list is empty.
         */
        public Task GetMinimum()
        {
            if (taskList.Count == 0)
            {
                throw new InvalidOperationException("The critical queue is empty!") ;
            }

            Task minimum = taskList[0] ;

            for (int i = 1 ; i < taskList.Count ; i++)
            {
                if (taskList[i].MaximumStartTime < minimum.MaximumStartTime)
                {
                    minimum = taskList[i] ;
                }
            }

            return minimum ;
        }
    
        /**
         * Removes the specified task from the list.
         */
        public void Remove(Task task)
        {
            taskList.Remove(task) ;
        }
    }
}
