﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_APX

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The alternating position crossover operator implementation.
     *
     * @author Tomas Tunys.
     */
    public class CAlternatingPositionCrossover : CBaseCrossoverOperator, IConfigurationDependent
    {
        /** Marks already transfered gene with given allele value. */
        private bool[] m_usedAllele ;

        /** The chromosome pool that is used to acquire chromosome instances. */
        private IChromosomePool m_chromosomePool ;

        /** The chromosome size. */
        private int m_chromosomeSize ;

        /**
         * Creates new alternating crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         */
        public CAlternatingPositionCrossover(CConfiguration configuration, double crossoverRate) : base(configuration, crossoverRate)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_chromosomePool = configuration.getChromosomePool() ;
            m_chromosomeSize = configuration.getChromosomeSize() ;
            m_usedAllele     = new bool[m_chromosomeSize] ;
        }        

        /**
         * Transferes genes from dad and mom chromosome on the child in alternating
         * way, such that no two genes with the same allele value are transferred.
         * 
         * @param dad   The 1st parent chromosome.
         * @param mom   The 2nd parent chromosome.
         * @param child The resulting chromosome of the crossover.
         */
        public void doCrossover(IChromosome dad, IChromosome mom, IChromosome child)
        {
            int i = 0, idad = 0, imom = 0 ;

            while (i < m_chromosomeSize)
            {
                if (idad < m_chromosomeSize)
                {
                    IGene gene = dad.getGene(idad++) ;

                    if (!m_usedAllele[(int)gene.getAllele()])
                    {
                        child.setGene(gene, i++) ;
                        m_usedAllele[(int)gene.getAllele()] = true ;
                    }
                }

                if (imom < m_chromosomeSize)
                {
                    IGene gene = mom.getGene(imom++) ;

                    if (!m_usedAllele[(int)gene.getAllele()])
                    {
                        child.setGene(gene, i++) ;
                        m_usedAllele[(int)gene.getAllele()] = true ;
                    }
                }
            }

            for (i = 0 ; i < m_chromosomeSize ; i++)
            {
                m_usedAllele[i] = false ;
            }
        }

        /**
         * Crossovers the specified couple of chromosomes in a operator-specific way
         * forming a new chromosomes that should be added to the specified population.
         * 
         * @param daddy      The first chromosome entering into the crossover.
         * @param mommy      The second chromosome entering into the crossover.
         * @param population The population for the next generation.
         */
        public override void apply(IChromosome daddy, IChromosome mommy, CPopulation population)
        {
            IChromosome son = m_chromosomePool.acquire() ;
            
            if (son != null)
            {
                son.setGenes(daddy.getGenes()) ;
            }
            else
            {
                son = (IChromosome)daddy.Clone() ;
            }

            IChromosome daughter = m_chromosomePool.acquire() ;
            
            if (daughter != null)
            {
                daughter.setGenes(mommy.getGenes()) ;
            }
            else
            {
                daughter = (IChromosome)mommy.Clone() ;
            }

            if (shouldPerformCrossover() && !daddy.Equals(mommy))
            {
                doCrossover(daddy, mommy, son) ;
                doCrossover(mommy, daddy, daughter) ;
            }

            population.addChromosome(son) ;
            population.addChromosome(daughter) ;
#if DEBUG_APX
            printCrossoverDebugInfo("Alternating-Position Crossover:", daddy, mommy, son, daughter) ;
#endif
        }
    }
}