﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

/**
 * Contains the implementation of a (generic) genetic algorithm that 
 * may be used to solve different kind of optimization problems.
 * It contains a lot of permutation-based crossover and mutation
 * operators, fitness scaling methods, and methods of selection.
 */
namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A base class for implementation of a genetic algorithm.
     */
    public abstract class CBaseGeneticAlgorithm
    {
        /** The configuration of this genetic algorithm. */
        private CConfiguration m_configuration ;

        /** The current population of chromosomes. */
        private CPopulation m_population ;

        /** The evolved population of chromosomes. */
        private CPopulation m_evolvedPopulation ;

        /** The generation counter. */
        private int m_generationCount ;

        /**
         * Creates and initializes the genetic algorithm.
         * 
         * @param configuration The configuration for this genetic 
         *                      algorithm.
         */
        protected CBaseGeneticAlgorithm(CConfiguration configuration)
        {
            setConfiguration(configuration) ;
        }

        /**
         * Sets up the configuration of this genetic algorithm.
         * 
         * @param configuration The settings of this genetic algorithm.
         * 
         * @throws NullReferenceException If configuration is <em>null</em>.
         */
        public void setConfiguration(CConfiguration configuration)
        {
            if (configuration == null) 
            {
                throw new NullReferenceException("The configuration must not be null!") ;
            }
            else
            {
                m_configuration = configuration ;
            }
        }

        /**
         * Returns the configuration of this genetic algorithm.
         * 
         * @return The configuration of this genetic algorithm.
         */
        public CConfiguration getConfiguration()
        {
            return m_configuration ;
        }

        /**
         * Creates a new initial population for the start of the genetic 
         * algorithm. The size of the population must agree with the value
         * set in the configuration, see <em>this.reset()</em> for explanation.
         * 
         * @param population The initial population to be filled up
         *                   with chromosomes.
         */
        public abstract void createInitialPopulation(CPopulation population) ;

        /**
         * Resets the genetic algorithm to initial state, which means
         * a new initial population is generated and the generation
         * counter is set to 0.
         * <p>
         * <strong>Note:</strong> This method must be called prior to
         *                        calling the <em>this.evaluate</em> method.
         * </p>
         * 
         * @throws InvalidOperationException if the size of the initial population
         *                                   created by calling createInitialPopulation
         *                                   does not agree with the size set in
         *                                   the configuration.
         */
        public void reset()
        {
            m_generationCount = 0 ;

            /* If called for the first time. */
            if (m_population == null)
            {
                m_population        = new CPopulation(getConfiguration()) ;
                m_evolvedPopulation = new CPopulation(getConfiguration()) ;

                /* TODO: This atribute might be set through the configuration!?. */
                m_population.setMaximumSize(getConfiguration().getPopulationSize()) ;
                m_evolvedPopulation.setMaximumSize(getConfiguration().getPopulationSize()) ;
            }
            else
            {
                m_population.clear() ;
                m_evolvedPopulation.clear() ;
            }

            if (getConfiguration().getChromosomePool() != null)
            {
                getConfiguration().getChromosomePool().clear() ;
                getConfiguration().getChromosomePool().setMaximumSize(2 * getConfiguration().getPopulationSize()) ;
            }
            else
            {
                throw new InvalidOperationException("The chromosome pool cannot be null!") ;
            }


            getConfiguration().notifyConfigurationDependentComponents() ;

            /* Creates a new initial population. */
            createInitialPopulation(m_population) ;

            /* Validates the size of the initial population. */
            if (m_population.getSize() != getConfiguration().getPopulationSize())
            {
                throw new InvalidOperationException("The size of the initial population does not agree" +
                                                    "with the value set in the configuration!") ;
            }
            
            /* Calculates the chromosomes' fitness scores. */
            getConfiguration().getFitnessEvaluator().evaluate(m_population) ;

            /* Scales the "raw" fitness values. */
            getConfiguration().getFitnessScaler().scale(m_population) ;
        }

        /**
         * Performs single step of evolution, and returns the best chromosome
         * of the evolved population.
         * 
         * @return the best chromosome of the evolved population.
         */
        public IChromosome evolve()
        {
            /* Uses the breeder to evolve the population. */
            getConfiguration().getBreeder().breed(m_population, m_evolvedPopulation) ;

            m_population.clear() ;

            CPopulation swapee  = m_population ;
            m_population        = m_evolvedPopulation ;
            m_evolvedPopulation = swapee ;

            ++m_generationCount ;

             /* Calculates the chromosomes' fitness scores. */
            getConfiguration().getFitnessEvaluator().evaluate(m_population) ;

            /* Scales the "raw" fitness values. */
            getConfiguration().getFitnessScaler().scale(m_population) ;
 
            return getFittestChromosome() ;
        }

        /**
         * Performs <em>count</em> number of evoluations, and then returns 
         * the best chromosome of the last evolved population.
         * 
         * @param count The number of evolutions to be performed.
         * 
         * @return the best chromosome of the evolved population.
         */
        public IChromosome evolve(int count)
        {
            for (int i = 0 ; i < count ; i++)
            {
                evolve() ;
            }

            return getFittestChromosome() ;
        }

        /**
         * Returns the fittest chromosome of the current population, i.e. 
         * the best current solution to the problem the algorithm
         * is supposed to solve.
         * 
         * @return The fittest chromosome of the current population.
         */
        public IChromosome getFittestChromosome()
        {
            return m_population.getFittestChromosome() ;
        }

        /**
         * Returns the generation number of the current population.
         * 
         * @return The generation number of the current population.
         */
        public int getGenerationNumber()
        {
            return m_generationCount ;
        }
    }
}
