﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A base class for implementation of a natural selector.
     */
    public abstract class CBaseSelectionOperator : CGeneticOperator, ISelectionOperator
    {
        /**
         * Creates and initializes the base selection operator.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this selector.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */
        protected CBaseSelectionOperator(CConfiguration configuration) : base(configuration)
        {
        }

        /**
         * Selects chromosomes from the population that will continue to the next
         * stage of the genetic algorithm: reproduction phase. This selection 
         * should be guided by the fitness values of the chromosomes, but that 
         * is up to the implementation of the selector.
         * 
         * @param population    The population the chromosome will be selected from.
         * @param matingPool    The mating pool the chromosome will be inserted into.
         * @param toSelectCount The total number of chromosomes to select.
         */
        public abstract void select(CPopulation population, CPopulation matingPool, int toSelectCount) ;

        /**
         * Enforces positive fitness scores of the chromosomes in the specified population
         * by adding the largest negative value to fitness scores of all the individuals of
         * the population.
         * 
         * @param population The concerned individuals.
         */
        public void enforcePositiveFitnessScores(CPopulation population)
        {
            double minimumFitnessValue = population.getWorstFitChromosome().getFitnessValue() ;

            if (minimumFitnessValue < 0.0)
            {                
                for (int i = 0 ; i < population.getSize() ; i++)
                {
                    population[i].setFitnessValue(population[i].getFitnessValue() - minimumFitnessValue) ;
                }
            }
        }
    }
}
