﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The basic implementation of the IChromosomePool interface.
     * The pool is represented by a dynamic list of chromosomes.
     */
    class CChromosomePool : IChromosomePool
    {
        /* The list of chromosome instances. */
        private List<IChromosome> m_chromosomePool ;

        /**
         * Creates a new chromosome pool with the maximum
         * size of 64 instances.
         */
        public CChromosomePool() : this(64)
        {
        }

        /**
         * Creates a new chromosome pool with the specified
         * maximum size.
         */
        public CChromosomePool(int size)
        {
            m_chromosomePool = new List<IChromosome>() ;

            setMaximumSize(size) ;
        }

        /**
         * Returns a chromosome from the pool. If the pool
         * is empty, null is returned.
         * 
         * @return A chromosome, or null, if the pool is
         *         empty.
         */
        public IChromosome acquire()
        {
            IChromosome chromosome = null ;
            
            if (m_chromosomePool.Count > 0)
            {
                chromosome = m_chromosomePool[m_chromosomePool.Count - 1] ;
                
                m_chromosomePool.RemoveAt(m_chromosomePool.Count - 1) ;
            }

            return chromosome ;
        }

        /**
         * Puts the specified chromosome into the pool so
         * it may be reused. 
         * 
         * @param chromosome The chromosome to be added
         *                   into the pool.
         */
        public void release(IChromosome chromosome)
        {
            if (m_chromosomePool.Count < m_chromosomePool.Capacity)
            {
                chromosome.dispose() ;
                m_chromosomePool.Add(chromosome) ;
            }
        }

        /**
         * Sets the capacity of the pool. Any chromosome
         * added when the pool is at its maximum capacity
         * is discarded.
         * 
         * @param size The size of the pool.
         */
        public void setMaximumSize(int size)
        {
            if (m_chromosomePool.Count > size)
            {
                m_chromosomePool.RemoveRange(size, m_chromosomePool.Count - size) ;
            }
            m_chromosomePool.Capacity = size + 1 ;
        }

        /**
         * Empties the chromosome pool.
         */
        public void clear()
        {
            m_chromosomePool.Clear() ;
        }
    }
}
