﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_CX

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The cycle crossover operator implementation.
     *
     * @author Tomas Tunys.
     */
    public class CCycleCrossover : CBaseCrossoverOperator, IConfigurationDependent
    {
        private int[] m_inverse ;

        /** The chromosome pool that is used to acquire chromosome instances. */
        private IChromosomePool m_chromosomePool ;

        /** The chromosome size. */
        private int m_chromosomeSize ;

        /**
         * Creates new cycle crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         */
        public CCycleCrossover(CConfiguration configuration, double crossoverRate) : base(configuration, crossoverRate)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_chromosomePool = configuration.getChromosomePool() ;
            m_chromosomeSize = configuration.getChromosomeSize() ;
            m_inverse        = new int[m_chromosomeSize] ;
        }  

        /**
         * Crossovers the specified couple of chromosomes in a operator-specific way
         * forming a new chromosomes that should be added to the specified population.
         * 
         * @param daddy      The first chromosome entering into the crossover.
         * @param mommy      The second chromosome entering into the crossover.
         * @param population The population for the next generation.
         */
        public override void apply(IChromosome daddy, IChromosome mommy, CPopulation population)
        {
            IChromosome son = m_chromosomePool.acquire() ;
            
            if (son != null)
            {
                son.setGenes(daddy.getGenes()) ;
            }
            else
            {
                son = (IChromosome)daddy.Clone() ;
            }

            IChromosome daughter = m_chromosomePool.acquire() ;
            
            if (daughter != null)
            {
                daughter.setGenes(mommy.getGenes()) ;
            }
            else
            {
                daughter = (IChromosome)mommy.Clone() ;
            }

            if (shouldPerformCrossover() && !daddy.Equals(mommy))
            {
                for(int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    m_inverse[(int)son.getGene(i).getAllele()] = i ;
                }

                int k = 0 ; 
            
                do
                {
                    swapGenes(son, daughter, k, k) ;

                    k = m_inverse[(int)son.getGene(k).getAllele()] ;
                }
                while (k != 0) ;
            }

            population.addChromosome(son) ;
            population.addChromosome(daughter) ;
#if DEBUG_CX
            printCrossoverDebugInfo("Cycle Crossover:", daddy, mommy, son, daughter) ;
#endif
        }
    }
}
