﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A default implementation of IBreeder interface. Instances of a default
     * breeder are practically useful for most of the problems.
     */
    public class CDefaultBreeder : CBaseBreeder, IConfigurationDependent
    {
        /** The pool for chromosomes for further reproduction. */
        private CPopulation m_matingPool ;

        /** The priority queue used to get fittest chromosomes for the next generation. */
        private CPriorityQueue<IChromosome> m_chooseFittestQueue ;

        /** The number of fittest chromosomes to copy into the new generation. */
        private int m_numberOfFittestToCopy ;

        /**
         * Creates a new default implementation of a breeder.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this breeder.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */ 
        public CDefaultBreeder(CConfiguration configuration) : base(configuration)
        {   
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this breeder.
         */
        public void notify(CConfiguration configuration)
        {
            m_matingPool = new CPopulation(configuration, configuration.getPopulationSize()) ;
            
            m_numberOfFittestToCopy = configuration.getNumberOfFittestToCopy() ;

            if (m_numberOfFittestToCopy > 0)
            {
                m_chooseFittestQueue = new CPriorityQueue<IChromosome>(configuration.getFitnessComparer(), m_numberOfFittestToCopy, true) ;
            }
        }

        /**
         * Evolves the specified population of chromosomes, by execution of
         * all genetic operators: selectors, crossover, and mutation operators 
         * that are currently set up in the configuration of the algorithm
         * that has been specified during initialization of the breeder.
         * <p>
         * <strong>Note:</strong> The fitness scores of the population is expected 
         * to be assigned prior to calling this method.
         * </p>
         *
         * @param population        The population to be evolved.
         * @param evolvedPopulation A new evolved population.
         */   
        public override void breed(CPopulation population, CPopulation evolvedPopulation)
        {
            /* No chromosome has been selected for the new generation, yet. */
            for (int i = 0 ; i < population.getSize() ; i++)
            {
                population[i].setSelectedForNextGeneration(false) ;
            }

            /* Selects chromosomes that will get chance to reproduce themselves. */
            getConfiguration().getNaturalSelector().select(population, m_matingPool, population.getSize()) ;

            List<ICrossoverOperator> crossoverOperators = getConfiguration().getCrossoverOperators() ;
            int                      crossoverIndex     = 0 ;
            int                      crossoverCount     = crossoverOperators.Count ;
            
            /* Applies crossover operators one by one until the output population is full, leaving place for fittest chromosomes. */
            while (evolvedPopulation.getSize() < population.getSize() - m_numberOfFittestToCopy)
            {
                IChromosome daddy = m_matingPool[getConfiguration().getRandomGenerator().Next(m_matingPool.getSize())] ;
                IChromosome mommy = m_matingPool[getConfiguration().getRandomGenerator().Next(m_matingPool.getSize())] ;

                crossoverOperators[crossoverIndex].apply(daddy, mommy, evolvedPopulation) ;

                if (++crossoverIndex == crossoverCount)
                {
                    crossoverIndex = 0 ;
                }
            }

            /* Empties the mating pool. */
            m_matingPool.clear() ;

            List<IMutationOperator>  mutationOperators = getConfiguration().getMutationOperators() ;
            int                      mutationIndex     = 0 ;
            int                      mutationCount     = mutationOperators.Count ; 
            
            /* Applies mutation operators one by one for all reproduces chromosomes, except for <m_numberOfFittestToCopy>
             * number of fittest chromosomes that have been copied. */
            for (int i = 0 ; i < evolvedPopulation.getSize() ; i++)
            {
                mutationOperators[mutationIndex].apply(evolvedPopulation[i]) ;

                if (++mutationIndex == mutationCount)
                {
                    mutationIndex = 0 ;
                }
            }

            /* If elitism is turned on - insert the given amount into the next generation rightaway. */
            if (m_numberOfFittestToCopy > 0)
            {
                if (population.isSorted())
                {
                    IEnumerator<IChromosome> enumerator = population.getFittestChromosomes(m_numberOfFittestToCopy) ;

                    while (enumerator.MoveNext())
                    {
                        evolvedPopulation.addChromosome(enumerator.Current) ;
                    }
                }
                else
                {
                    for (int i = 0 ; i < population.getSize() ; i++)
                    {
                        m_chooseFittestQueue.push(population[i]) ;
                    }

                    foreach (IChromosome chromosome in m_chooseFittestQueue.elements())
                    {
                        evolvedPopulation.addChromosome(chromosome) ;
                    }
                }
            }

            for (int i = 0 ; i < evolvedPopulation.getSize() ; i++)
            {
                evolvedPopulation[i].setSelectedForNextGeneration(true) ;
            }

            IChromosomePool chromosomePool = getConfiguration().getChromosomePool() ;

            for (int i = 0 ; i < population.getSize() ; i++)
            {
                if (!population[i].isSelectedForNextGeneration())
                {
                    chromosomePool.release(population[i]) ;
                }
            }
        }
    }
}