﻿using System ;
using System.Linq ;
using System.Text ;
using System.Collections.Generic ;
using System.Globalization ;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The serializer for the implemented finess scalers.
     */
    class CFitnessScalerSerializer
    {
        /**
         * Returns the System.String representing the serialization
         * of the specified fitness scaler.
         * 
         * @param scaler The scaler to be serialized.
         * 
         * @return The string representing the serialization of the 
         *         specified fitness scaler.
         * 
         * @throws ArgumentException If scaler is null or unknown
         *                           by this serializer.
         */
        public static string serialize(IFitnessScaler scaler)
        {
            if (scaler is CNullFitnessScaler)
            {
                return "NONE" ;
            }
            else if (scaler is CRankScaling)
            {
                return "RANK" ;
            }
            else if (scaler is CSigmaScaling)
            {
                return "SIGMA" ;
            }
            else if (scaler is CBoltzmannScaling)
            {
                CBoltzmannScaling s = (CBoltzmannScaling)scaler ;

                return String.Format(CultureInfo.InvariantCulture, "BOLTZMANN {0:0.####} {1:0.####} {2:0.####}", s.getStartTemperature(), s.getMinimumTemperature(), s.getCoolingStepSize()) ;
            }
            else
            {
                throw new ArgumentException("Uknown or null scaler!") ;
            }
        }

        /**
         * Creates a fitness scaler from the specified string.
         * 
         * @param serialization The fitness scaler serialization to be
         *                      deserialized.
         *                      
         * @return The deserialized fitness scaler.
         *                      
         * @throws ArgumentException If the serialization is invalid,
         *                           or represents no fitness scaler.
         */
        public static IFitnessScaler deserialize(string serialization)
        {
            if (String.Compare(serialization, "none", true) == 0)
            {
                return new CNullFitnessScaler() ;
            }
            else if (String.Compare(serialization, "rank", true) == 0)
            {
                return new CRankScaling() ;
            }
            else if (String.Compare(serialization, "sigma", true) == 0)
            {
                return new CSigmaScaling() ;
            }
            else if (serialization.StartsWith("boltzmann", true, null))
            {
                String[] parameters = serialization.Split(' ', '\t') ;

                if (parameters.Length != 4)
                {
                    throw new ArgumentException("Boltzmann scaler parameters missing!") ;
                }

                return new CBoltzmannScaling(double.Parse(parameters[1], CultureInfo.InvariantCulture), double.Parse(parameters[2], CultureInfo.InvariantCulture), double.Parse(parameters[3], CultureInfo.InvariantCulture)) ;
            }
            else
            {
                throw new ArgumentException("Unknonw scaler to be deserialized!") ;
            }
        }
    }
}
