﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_IM

using System ;
using System.Linq ;
using System.Text ;
using System.Collections.Generic ;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The insertion mutation operator. It randomly chooses a gene within
     * a chromosome with uniform probability, and then the gene is displaced
     * and reinserted back into randomly chosen position.
     * 
     * @author Tomas Tunys
     */
    public class CInsertionMutation : CBaseMutationOperator, IConfigurationDependent
    {
        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /**
         * Creates new insertion mutation operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param mutationRate  The mutation rate.
         */
        public CInsertionMutation(CConfiguration configuration, double mutationRate) : base(configuration, mutationRate)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_randomizer = configuration.getRandomGenerator() ;
        }

        /**
         * Mutates the specified chromosome in a operator-specific way forming 
         * a new chromosome, which will then go to the next generation.
         * 
         * @param chromosome The chromosome to be mutated.
         */
        public override void apply(IChromosome chromosome)
        {
            if (shouldPerformMutation())
            {
                /* Randomly choses two points. */
                int i = m_randomizer.Next(chromosome.getSize()) ;
                int j = m_randomizer.Next(chromosome.getSize()) ;

                if (i != j)
                {
                    int step = (i < j ? 1 : -1) ;

                    IGene gene = chromosome.getGene(i) ;
#if DEBUG_IM
                    Console.WriteLine("Insertion Mutation: " + (i + 1) + " -> " + (j + 1)) ;
                    printOutGenome(chromosome, (int)Math.Ceiling(Math.Log10(chromosome.getSize())), true) ;
#endif
                    /* Displaces the chosen gene... */
                    while (i != j)
                    {
                        chromosome.setGene(chromosome.getGene(i + step), i) ;
                        i += step ;
                    }

                    /* ...and reinserts it. */
                    chromosome.setGene(gene, j) ;
#if DEBUG_IM
                    printOutGenome(chromosome, (int)Math.Ceiling(Math.Log10(chromosome.getSize())), false) ;
#endif
                }
            }
        }
    }
}
