﻿using System ;
using System.Linq ;
using System.Text ;
using System.Globalization ;
using System.Collections.Generic ;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A serializer for mutation operators.
     */
    class CMutationOperatorSerializer
    {
        /**
         * Returns the System.String representing the serialization
         * of the specified mutation operator.
         * 
         * @param mutation The operator to be serialized.
         * 
         * @return The string representing the serialization of the
         *         specified mutation operator.
         * 
         * @throws ArgumentException If mutation is null or unknown
         *                           by this serializer.
         */
        public static string serialize(IMutationOperator mutation)
        {
            if (mutation is CDisplacementMutation)
            {
                return String.Format(CultureInfo.InvariantCulture, "DM {0:0.####}", ((CDisplacementMutation)mutation).getMutationRate()) ;
            }
            else if (mutation is CInsertionMutation)
            {
                return String.Format(CultureInfo.InvariantCulture, "IM {0:0.####}", ((CInsertionMutation)mutation).getMutationRate()) ;
            }
            else if (mutation is CInversionMutation)
            {
                return String.Format(CultureInfo.InvariantCulture, "IVM {0:0.####}", ((CInversionMutation)mutation).getMutationRate()) ;
            }
            else if (mutation is CExchangeMutation)
            {
                return String.Format(CultureInfo.InvariantCulture, "EM {0:0.####}", ((CExchangeMutation)mutation).getMutationRate()) ;
            }
            else if (mutation is CScrambleMutation)
            {
                return String.Format(CultureInfo.InvariantCulture, "SM {0:0.####}", ((CScrambleMutation)mutation).getMutationRate()) ;
            }
            else if (mutation is CDisplacedInversionMutation)
            {
                return String.Format(CultureInfo.InvariantCulture, "DIVM {0:0.####}", ((CDisplacedInversionMutation)mutation).getMutationRate()) ;
            }
            else
            {
                throw new ArgumentException("An unknown (or null) mutation operator!") ;
            }
        }

        /**
         * Creates a mutation operator from the specified string.
         * 
         * @param serialization The operator's serialization to be
         *                      deserialized.
         * @param configuration The configuration of the genetic
         *                      algorithm that will use the mutation
         *                      operator.         
         *                      
         * @return The deserialized mutation operator.
         *                      
         * @throws ArgumentException If the serialization is invalid,
         *                           or represents no mutation operator.
         */
        public static IMutationOperator deserialize(string serialization, CConfiguration configuration)
        {
            string[] parameters = serialization.Split(' ', '\t') ;

            if (serialization.StartsWith("DM", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for DM operator!") ;
                }

                return new CDisplacementMutation(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("IM", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for IM operator!") ;
                }

                return new CInsertionMutation(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("IVM", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for IVM operator!") ;
                }

                return new CInversionMutation(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("EM", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for EM operator!") ;
                }

                return new CExchangeMutation(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("SM", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for SM operator!") ;
                }

                return new CScrambleMutation(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("DIVM", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for DIVM operator!") ;
                }

                return new CDisplacedInversionMutation(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else
            {
                throw new ArgumentException("Unknown selector operator!") ;
            }
        }
    }
}
