﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_OBX

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The order-based crossover operator implementation.
     *
     * @author Tomas Tunys.
     */
    public class COrderBasedCrossover : CBaseCrossoverOperator, IConfigurationDependent
    {
        /** The order-imposed genes of the daddy chromosome. */
        private IGene[] m_orderedDadGenes ;

        /** The order-imposed genes of the mommy chromosome. */
        private IGene[] m_orderedMomGenes ;

        /** Marks the allele values of the daddy's order-imposed genes. */
        private bool[] m_orderedDadAlleles ;

        /** Marks the allele values of the mommy's order-imposed genes. */
        private bool[] m_orderedMomAlleles ;

        /** The ratio of information transfered from each of the parents on a child. */
        private double  m_ration ;

        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /** The chromosome pool that is used to acquire chromosome instances. */
        private IChromosomePool m_chromosomePool ;

        /** The chromosome size. */
        private int m_chromosomeSize ;

        /**
         * Creates new order-based crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         */
        public COrderBasedCrossover(CConfiguration configuration, double crossoverRate) : this(configuration, crossoverRate, 0.5)
        {
        }

        /**
         * Creates new order-based crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         * @param ratio         The ratio of information transfered from each of 
         *                      the parents on a child.
         *                      
         * @throws ArgumentOutOfRangeException If the ration is not in range 0.0 to 1.0.
         */
        public COrderBasedCrossover(CConfiguration configuration, double crossoverRate, double ratio) : base(configuration, crossoverRate)
        {
            if (ratio < 0.0 || ratio > 1.0)
            {
                throw new System.ArgumentOutOfRangeException("The ratio must to be between 0.0 and 1.0 inclusive!") ;
            }

            m_ration = ratio ;
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_randomizer        = configuration.getRandomGenerator() ;
            m_chromosomePool    = configuration.getChromosomePool() ;
            m_chromosomeSize    = configuration.getChromosomeSize() ;
            m_orderedDadGenes   = new IGene[configuration.getChromosomeSize()] ;
            m_orderedMomGenes   = new IGene[configuration.getChromosomeSize()] ;
            m_orderedDadAlleles = new bool[configuration.getChromosomeSize()] ;
            m_orderedMomAlleles = new bool[configuration.getChromosomeSize()] ;
        }

        /**
         * Returns The ratio of information transfered from each
         * of the parents on a child.
         * 
         * @return The ratio of information transfered from each 
         *         of the parents on a child.
         */
        public double getCrossoverRatio()
        {
            return m_ration ;
        }

        /**
         * Crossovers the specified couple of chromosomes in a operator-specific way
         * forming a new chromosomes that should be added to the specified population.
         * 
         * @param daddy      The first chromosome entering into the crossover.
         * @param mommy      The second chromosome entering into the crossover.
         * @param population The population for the next generation.
         */
        public override void apply(IChromosome daddy, IChromosome mommy, CPopulation population)
        {
            IChromosome son = m_chromosomePool.acquire() ;
            
            if (son != null)
            {
                son.setGenes(daddy.getGenes()) ;
            }
            else
            {
                son = (IChromosome)daddy.Clone() ;
            }

            IChromosome daughter = m_chromosomePool.acquire() ;
            
            if (daughter != null)
            {
                daughter.setGenes(mommy.getGenes()) ;
            }
            else
            {
                daughter = (IChromosome)mommy.Clone() ;
            }

            if (shouldPerformCrossover() && !daddy.Equals(mommy))
            {
#if DEBUG_OBX
                /* Stores the position of genes whose order's going to be
                 * imposed on the childeren.
                 */
                string orderedLoci = "" ;
#endif
                /* The number of fixed gene positions. */
                int numberOfFixedGenes = 0 ;

                /* Randomly selects parent genes whose order is going 
                 * to be imposed on the childeren.
                 */
                for(int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    IGene dadGene = daddy.getGene(i) ;
                    IGene momGene = mommy.getGene(i) ;

                    if  (m_randomizer.NextDouble() < m_ration)
                    {
                        m_orderedDadAlleles[(int)dadGene.getAllele()] = true ;
                        m_orderedMomAlleles[(int)momGene.getAllele()] = true ;

                        m_orderedDadGenes[numberOfFixedGenes] = dadGene ;
                        m_orderedMomGenes[numberOfFixedGenes] = momGene ;

                        ++numberOfFixedGenes ;
#if DEBUG_OBX
                        orderedLoci += " " + (i + 1).ToString() ;
#endif
                    }
                }

                /* Imposes the order of the selected daddy genes on the daughter
                 * and similarily with mommy and son chromosomes.
                 */
                for (int i = 0, idad = 0, imom = 0 ; i < m_chromosomeSize ; i++)
                {
                    int daughterAllele = (int)daughter.getGene(i).getAllele() ;

                    if (m_orderedDadAlleles[daughterAllele])
                    {
                        if ((int)m_orderedDadGenes[idad].getAllele() != daughterAllele)
                        {
                            daughter.setGene(m_orderedDadGenes[idad], i) ;
                        }

                        ++idad ;
                    }

                    int sonAllele = (int)son.getGene(i).getAllele() ;

                    if (m_orderedMomAlleles[sonAllele])
                    {
                        if ((int)m_orderedMomGenes[imom].getAllele() != sonAllele)
                        {
                            son.setGene(m_orderedMomGenes[imom], i) ;
                        }

                        ++imom ;
                    }
                }

                /* Unmarks the previously marked allele values. */
                for(int i = 0 ; i < numberOfFixedGenes ; i++)
                {
                    m_orderedDadAlleles[(int)m_orderedDadGenes[i].getAllele()] = false ;
                    m_orderedMomAlleles[(int)m_orderedMomGenes[i].getAllele()] = false ;
                }
            }

            population.addChromosome(son) ;
            population.addChromosome(daughter) ;
#if DEBUG_OBX
            printCrossoverDebugInfo("Order-Based Crossover:" + orderedLoci, daddy, mommy, son, daughter) ;
#endif
        }
    }
}
