﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_OX

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The order crossover operator implementation.
     *
     * @author Tomas Tunys.
     */
    public class COrderCrossover: CBaseCrossoverOperator, IConfigurationDependent
    {
        /** The alleles values of order-fixed genes. */
        private bool[] m_ordered ;

        /** The ratio of information transfered from each of the parents on a child. */
        private double m_ration ;

        /** The minimum span of the randomly chosen section of a chromosome. */
        private int m_minSpan ;

        /** The maximum span of the randomly chosen section of a chromosome. */
        private int m_maxSpan ;

        /** The chromosome pool that is used to acquire chromosome instances. */
        private IChromosomePool m_chromosomePool ;

        /** The chromosome size. */
        private int m_chromosomeSize ;

        /**
         * Creates new order crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         */
        public COrderCrossover(CConfiguration configuration, double crossoverRate) : this(configuration, crossoverRate, 0.5)
        {
        }

        /**
         * Creates new order crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         * @param ratio         The ratio of information transfered from each of 
         *                      the parents on a child.
         *                      
         * @throws ArgumentOutOfRangeException If the ration is not in range 0.0 to 1.0.
         */
        public COrderCrossover(CConfiguration configuration, double crossoverRate, double ratio) : base(configuration, crossoverRate)
        {           
            if (ratio < 0.0 || ratio > 1.0)
            {
                throw new System.ArgumentOutOfRangeException("The ratio must to be between 0.0 and 1.0 inclusive!") ;
            }

            m_ration = ratio ;
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_chromosomePool = configuration.getChromosomePool() ;
            m_chromosomeSize = configuration.getChromosomeSize() ;
            m_ordered        = new bool[m_chromosomeSize] ;
            m_maxSpan        = m_chromosomeSize ;
            m_minSpan        = System.Math.Min(2, m_maxSpan) ;
        }

        /**
         * Returns The ratio of information transfered from each
         * of the parents on a child.
         * 
         * @return The ratio of information transfered from each 
         *         of the parents on a child.
         */
        public double getCrossoverRatio()
        {
            return m_ration ;
        }

        /**
         * Completes the crossover. The child chromosome is a copy of the second
         * parent different from the specified parent. Genes are transferred from
         * the parent onto child in permutation consistent way.
         * 
         * @param child The resulting child.
         * @param parent The parent from which genes are passed on the child.
         * @param start  The start of the order-fixed section of the chromosome.
         * @param span   The span of the order-fixed section of the chromosome.
         */
        private void doCrossover(IChromosome child, IChromosome parent, int start, int span)
        {
            /* Marks the allele values of genes that should NOT be transfered from the parent. */
            for (int i = 0 ; i < span ; i++)
            {
                m_ordered[(int)child.getGene(start + i).getAllele()] = true ;
            }

            int ichild  = (start + span) % m_chromosomeSize ;
            int iparent = (start + span) % m_chromosomeSize ;

            /* Until the genes transfer is complete. */
            while (ichild != start)
            {
                /* A gene for transfer. */
                IGene gene = parent.getGene(iparent) ;

                /* If the gene with this allele value is not in the order-fixed section... */
                if (!m_ordered[(int)gene.getAllele()])
                {
                    /* ... the gene is passed on the child. */
                    child.setGene(gene, ichild) ;

                    if (++ichild == m_chromosomeSize)
                    {
                        ichild = 0 ;
                    }
                }
                /* ...otherwise it is ignored. */

                if (++iparent == m_chromosomeSize)
                {
                    iparent = 0 ;
                }
            }

            /** Unmarks previously marked allele values. */
            for (int i = 0 ; i < span ; i++)
            {
                m_ordered[(int)child.getGene(start + i).getAllele()] = false ;
            }
        }

        /**
         * Crossovers the specified couple of chromosomes in a operator-specific way
         * forming a new chromosomes that should be added to the specified population.
         * 
         * @param daddy      The first chromosome entering into the crossover.
         * @param mommy      The second chromosome entering into the crossover.
         * @param population The population for the next generation.
         */
        public override void apply(IChromosome daddy, IChromosome mommy, CPopulation population)
        {
            IChromosome son = m_chromosomePool.acquire() ;
            
            if (son != null)
            {
                son.setGenes(daddy.getGenes()) ;
            }
            else
            {
                son = (IChromosome)daddy.Clone() ;
            }

            IChromosome daughter = m_chromosomePool.acquire() ;
            
            if (daughter != null)
            {
                daughter.setGenes(mommy.getGenes()) ;
            }
            else
            {
                daughter = (IChromosome)mommy.Clone() ;
            }

            if (shouldPerformCrossover() && !daddy.Equals(mommy))
            {
                int start, span ;

                /* Randomly chooses order-fixed section of a chromosome. */
                getChromosomeSection(out start, out span, m_minSpan, m_maxSpan) ;

                doCrossover(son, mommy, start, span) ;
                doCrossover(daughter, daddy, start, span) ;
            }

            population.addChromosome(son) ;
            population.addChromosome(daughter) ;

#if DEBUG_OX
            printCrossoverDebugInfo("Order Crossover: " + (start + 1) + " - " + (start + span), daddy, mommy, son, daughter) ;
#endif
        }
    }
}
