﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_PBX

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The postion-based crossover operator implementation.
     *
     * @author Tomas Tunys.
     */
    public class CPositionBasedCrossover : CBaseCrossoverOperator, IConfigurationDependent
    {
        /** The allele values of daddy genes whose position is imposed in the child chromosome. */
        private bool[] m_fixedDadAlleles ;

        /** The allele values of mommy genes whose position is imposed in the child chromosome. */
        private bool[] m_fixedMomAlleles ;

        /** The position-imposed chromosome positions. */
        private bool[] m_fixedLoci ;
        
        /** The ratio of information transfered from each of the parents on a child. */
        private double  m_ration ;

        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /** The chromosome pool that is used to acquire chromosome instances. */
        private IChromosomePool m_chromosomePool ;

        /** The chromosome size. */
        private int m_chromosomeSize ;

        /**
         * Creates new position-based crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         */
        public CPositionBasedCrossover(CConfiguration configuration, double crossoverRate) : this(configuration, crossoverRate, 0.5)
        {
        }

        /**
         * Creates new position-based crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         * @param ratio         The ratio of information transfered from each of 
         *                      the parents on a child.
         *                      
         * @throws ArgumentOutOfRangeException If the ration is not in range 0.0 to 1.0.
         */
        public CPositionBasedCrossover(CConfiguration configuration, double crossoverRate, double ratio) : base(configuration, crossoverRate)
        {
            if (ratio < 0.0 || ratio > 1.0)
            {
                throw new System.ArgumentOutOfRangeException("The ratio must to be between 0.0 and 1.0 inclusive!") ;
            }

            m_ration = ratio ;
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_randomizer      = configuration.getRandomGenerator() ;
            m_chromosomePool  = configuration.getChromosomePool() ;
            m_chromosomeSize  = configuration.getChromosomeSize() ;
            m_fixedDadAlleles = new bool[m_chromosomeSize] ;
            m_fixedMomAlleles = new bool[m_chromosomeSize] ;
            m_fixedLoci       = new bool[m_chromosomeSize] ;
        }

        /**
         * Returns The ratio of information transfered from each
         * of the parents on a child.
         * 
         * @return The ratio of information transfered from each 
         *         of the parents on a child.
         */
        public double getCrossoverRatio()
        {
            return m_ration ;
        }

        /**
         * Crossovers the specified couple of chromosomes in a operator-specific way
         * forming a new chromosomes that should be added to the specified population.
         * 
         * @param daddy      The first chromosome entering into the crossover.
         * @param mommy      The second chromosome entering into the crossover.
         * @param population The population for the next generation.
         */
        public override void apply(IChromosome daddy, IChromosome mommy, CPopulation population)
        {
            IChromosome son = m_chromosomePool.acquire() ;
            
            if (son != null)
            {
                son.setGenes(daddy.getGenes()) ;
            }
            else
            {
                son = (IChromosome)daddy.Clone() ;
            }

            IChromosome daughter = m_chromosomePool.acquire() ;
            
            if (daughter != null)
            {
                daughter.setGenes(mommy.getGenes()) ;
            }
            else
            {
                daughter = (IChromosome)mommy.Clone() ;
            }

            if (shouldPerformCrossover() && !daddy.Equals(mommy))
            {
#if DEBUG_PBX
                /* Stores the fixed positions for genes. */
                string fixedPositions = "" ;
#endif
                /* Randomly chooses fixed gene positions. */
                for(int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    m_fixedLoci[i] = (m_randomizer.NextDouble() < m_ration) ;
#if DEBUG_PBX
                    if (m_fixedLoci[i])
                    {
                        fixedPositions += " " + (i + 1).ToString() ;
                    }
#endif
                }

                /* Marks allele values of position-fixed genes for each of the parents. */
                for(int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    m_fixedDadAlleles[(int)daddy.getGene(i).getAllele()] = m_fixedLoci[i] ;
                    m_fixedMomAlleles[(int)mommy.getGene(i).getAllele()] = m_fixedLoci[i] ;
                }

                IGene[] dadGenes = daddy.getGenes() ;
                IGene[] momGenes = mommy.getGenes() ;

                /* Passes genes of the parents on the children. */
                for (int i = 0, idad = 0, imom = 0 ; i < m_chromosomeSize ; i++)
                {
                    if (m_fixedLoci[i])
                    {
                        son.setGene(dadGenes[i], i) ;
                        daughter.setGene(momGenes[i], i) ;
                    }
                    else
                    {
                        while (m_fixedMomAlleles[(int)dadGenes[idad].getAllele()])
                        {
                            ++idad ;
                        }

                        while (m_fixedDadAlleles[(int)momGenes[imom].getAllele()])
                        {
                            ++imom ;
                        }

                        son.setGene(momGenes[imom++], i) ;
                        daughter.setGene(dadGenes[idad++], i) ;
                    }
                }
            }

            population.addChromosome(son) ;
            population.addChromosome(daughter) ;
#if DEBUG_PBX
            printCrossoverDebugInfo("Position-Based Crossover:" + fixedPositions, daddy, mommy, son, daughter) ;
#endif
        }
    }
}
