﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.Serialization;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * Interface providing methods common to all type of chromosomes.
     * 
     * @author Tomas Tunys
     */
    public interface IChromosome : ICloneable, IEnumerable<IGene>
    {
        /**
         * Sets or gets the gene at the specified position.
         * 
         * @param index The zero-based index of the gene to get or set.
         * 
         * @return The gene at the specified position.
         */
        IGene this[int index] { get; set; }

        /**
         * Sets the gene at the specified position within the chromosome.
         * 
         * @param gene  The gene to be set into the chromosome.
         * @param index The zero-based position of the gene to be set up.
         */
        void setGene(IGene gene, int index) ;

        /**
         * Returns the gene at the given position within the chromosome.
         *
         * @param index The zero-based position of the gene value to be returned.
         * 
         * @return The gene at the given position.
         */
        IGene getGene(int index) ;

        /**
         * Sets up the genes of the chromosome.
         * 
         * @param genes The genes to be set in the chromosome.
         */   
        void setGenes(IGene[] genes) ;

        /**
         * Returns the set of genes that make up the chromosome.
         *
         * @return An array of genes the chromosome is build from.
         */
        IGene[] getGenes() ;

        /**
         * Returns the size (length) of the chromosome, i.e., the number of
         * genes the chromosome is build from.
         *
         * @return The number of genes the chromosome contains.
         */
        int getSize() ;

        /**
         * Sets the fitness value of the chromosome.
         * 
         * @param value A non-negative decimal value denoting how fit
         *              is the chromosome.           
         */
        void setFitnessValue(double value) ;

        /**
         * Returns the fitness value of the chromosome.
         *
         * @return A non-negative decimal value denoting how fit
         *         the chromosome is, or -1, if no fitness has been
         *         determined for it, yet.
         */
        double getFitnessValue() ;

        /**
         * Used to mark the chromosome "in use". Those that are not
         * selected will go to a chromosome pool.
         * 
         * @param selected If true, it marks the chromosome selected
         *                 for the next generation, otherwise, the
         *                 instance may go to a chromosome pool.
         */
        void setSelectedForNextGeneration(bool selected) ;

        /**
         * Returns true if this chromosome has been selected to go
         * to the next generation; otherwise, it returns false.
         * 
         * @return True if the chromosome goes to the next generation;
         *         otherwise, false.
         */
        bool isSelectedForNextGeneration() ;

        /**
         * This method is called when the chromosome is no longer
         * need it. This call gives the chromosome a chance to
         * release any resources it utilizes to be collected.
         */
        void dispose() ;


    }
}
