﻿using System;
using System.IO ;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MCScheduling.MixedCriticality.MIP ;
using MCScheduling.MixedCriticality.GA ;
using MCScheduling.MixedCriticality.SA ;
using MCScheduling.MixedCriticality.CEDF ;
using MCScheduling.MixedCriticality.DP ;
using System.Globalization;

namespace MCScheduling.MixedCriticality
{
    /**
     * The command line interface for the MCScheduling application. It
     * is the counterpart to the GUI that is normally display when
     * no command line arguments are fed to the application.
     */
    class CConsoleSolver
    {
        private const int NO_SOLVER   = -1 ;
        private const int MIP_SOLVER  = 0 ;
        private const int GA_SOLVER   = 1 ;
        private const int SA_SOLVER   = 2 ;
        private const int CEDF_SOLVER = 3 ;
        private const int DP_SOLVER   = 4 ;

        /** The type identificator of the mixed-criticality solver. */
        private int mSolverType = NO_SOLVER ;

        /** The mixed-criticality solver. */
        private CBaseMixedCriticalitySolver mSolver = null ;

        /** The name of the solver's configuration file. */
        private string mSolverConfigFileName = null ;
        
        /** The time limit in seconds. */
        private int mTimeLimit = -1 ;

        /** The iteration limit. */
        private int mIterationLimit = -1 ;
        
        /** The number of repetitions. */
        private int mRepetitionCount = -1 ;

        /** The name of the input file. */
        private string mInputFileName = null ;

        /** The name of the output file. */
        private string mOutputFileName = null ;

        /** The mixed-criticality instance to solve. */
        private CMixedCriticalityInstance mInstance = null ;

        /**
         * Creates the mixed-criticality solver that is executed
         * from a command line. 
         * 
         * @param arguments The command line arguments for the solver.
         */
        public CConsoleSolver(string[] arguments)
        {
            for (int i = 1 ; i < arguments.Length ; i += 2)
            {       
                if (String.Compare(arguments[i].Trim().ToLowerInvariant(), "-s") == 0)
                {
                    parseSolverType(arguments[i + 1]) ;
                }
                else if (String.Compare(arguments[i].Trim().ToLowerInvariant(), "-f") == 0)
                {
                    mSolverConfigFileName = arguments[i + 1].Trim().ToLowerInvariant() ;
                }
                else if (String.Compare(arguments[i].Trim().ToLowerInvariant(), "-t") == 0)
                {
                    mTimeLimit = int.Parse(arguments[i + 1].Trim()) ;
                }
                else if (String.Compare(arguments[i].Trim().ToLowerInvariant(), "-i") == 0)
                {
                    mIterationLimit = int.Parse(arguments[i + 1].Trim()) ;
                }
                else if (String.Compare(arguments[i].Trim().ToLowerInvariant(), "-c") == 0)
                {
                    mRepetitionCount = int.Parse(arguments[i + 1].Trim()) ;
                }
                else if (String.Compare(arguments[i].Trim().ToLowerInvariant(), "-in") == 0)
                {
                    mInputFileName = arguments[i + 1].Trim().ToLowerInvariant() ;
                }
                else if (String.Compare(arguments[i].Trim().ToLowerInvariant(), "-out") == 0)
                {
                    mOutputFileName = arguments[i + 1].Trim().ToLowerInvariant() ;
                }
                else
                {
                    throw new ArgumentException("Unexpected argument type: '" + arguments[i] + "'.") ;
                }
            }

            mInstance = CMixedCriticalityInstancesSerializer.deserialize(mInputFileName) ;
        }

        /**
         * Parses the type of the solver from the specified parameter
         * given from the command line.
         * 
         * @param solverParam The solver type parameter on the command line.
         */
        private void parseSolverType(string solverParam)
        {
            if (String.Compare(solverParam.Trim().ToLowerInvariant(), "mip") == 0)
            {
                mSolverType = MIP_SOLVER ;
            }
            else if (String.Compare(solverParam.Trim().ToLowerInvariant(), "ga") == 0)
            {
                mSolverType = GA_SOLVER ;
            }
            else if (String.Compare(solverParam.Trim().ToLowerInvariant(), "sa") == 0)
            {
                mSolverType = SA_SOLVER ;
            }
            else if (String.Compare(solverParam.Trim().ToLowerInvariant(), "cedf") == 0)
            {
                mSolverType = CEDF_SOLVER ;
            }
            else if (String.Compare(solverParam.Trim().ToLowerInvariant(), "dp") == 0)
            {
                mSolverType = DP_SOLVER ;
            }
            else
            {
                mSolverType = NO_SOLVER ;
            }
        }

        /**
         * Outputs a solution to the last solved MC instance into a file
         * that has been specified on the command line.
         * 
         * @param solutionStatus The status of the solution.
         */
        private void outputSolution(int solutionStatus)
        {
            using (StreamWriter fileWriter = new StreamWriter(mOutputFileName))
            {
                if (solutionStatus == CBaseMixedCriticalitySolver.Status.OPTIMAL || solutionStatus == CBaseMixedCriticalitySolver.Status.SUBOPTIMAL || solutionStatus == CBaseMixedCriticalitySolver.Status.HEURISTIC)
                {
                    fileWriter.WriteLine(String.Format(CultureInfo.InvariantCulture, "{0:0.##}", mInstance.getMakeSpan())) ;
                    fileWriter.WriteLine(String.Format(CultureInfo.InvariantCulture, "{0:0.##}", mInstance.getLateness())) ;
                    fileWriter.WriteLine(String.Format(CultureInfo.InvariantCulture, "{0:0.###}", mInstance.getCPUTime() / 1000.0)) ;
                }

                switch (solutionStatus)
                {
                case CBaseMixedCriticalitySolver.Status.OPTIMAL:
                    fileWriter.WriteLine("optimal") ;
                    break ;

                case CBaseMixedCriticalitySolver.Status.SUBOPTIMAL:
                    fileWriter.WriteLine("suboptimal") ;
                    break ;

                case CBaseMixedCriticalitySolver.Status.HEURISTIC:
                    fileWriter.WriteLine("heuristic") ;
                    break ;

                case CBaseMixedCriticalitySolver.Status.TIME_LIMIT:
                    fileWriter.WriteLine("time limit exceeded") ;
                    break ;

                case CBaseMixedCriticalitySolver.Status.ITERATION_LIMIT:
                    fileWriter.WriteLine("iteration limit exceeded") ;
                    break ;

                case CBaseMixedCriticalitySolver.Status.INFEASIBLE:
                    fileWriter.WriteLine("infeasible") ;
                    break ;

                case CBaseMixedCriticalitySolver.Status.INTERRUPTED:
                    fileWriter.WriteLine("interrupted") ;
                    break ;

                case CBaseMixedCriticalitySolver.Status.ERROR:
                    fileWriter.WriteLine("error") ;
                    break ;

                default:
                    fileWriter.WriteLine("unknown") ;
                    break ;
                }
            }
        }

        /**
         * Runs the selected solver on the given MC instance with the parameters
         * specified on the command line.
         */
        public void run() 
        {
            switch (mSolverType)
            {
            case MIP_SOLVER:
                mSolver = new CMixedIntegerProgrammingSolver() ;
                break ;

            case GA_SOLVER:
                mSolver = new CGeneticAlgorithmSolver(new CMixedCriticalityConfiguration(mSolverConfigFileName, 0)) ;
                break ;

            case SA_SOLVER:
                mSolver = new CSimulatedAnnealingSolver(new CSimulatedAnnealingConfiguration(mSolverConfigFileName)) ;
                break ;

            case CEDF_SOLVER:
                mSolver = new CClairvoyantEDFSolver() ;
                break ;

            case DP_SOLVER:
                mSolver = new CDPSolver() ;
                break ;

            case NO_SOLVER:
            default:
                throw new InvalidOperationException("No solver specified!") ;
            }

            mSolver.setTimeLimit(mTimeLimit >= 0 ? mTimeLimit : -1) ;

            mSolver.setIterationLimit(mIterationLimit >= 0 ? mIterationLimit : -1) ;

            System.Diagnostics.Stopwatch stopwatch = new System.Diagnostics.Stopwatch() ;

            if (mRepetitionCount > 1)
            {
                double makespan = 0.0 ;
                double lateness = 0.0 ;
                long   cpu      = 0 ;

                int unsolved = 0 ;

                stopwatch.Start() ;

                for (int i = 0 ; i < mRepetitionCount ; i++)
                {
                    stopwatch.Start() ;

                    mSolver.solve(mInstance) ;

                    stopwatch.Stop() ;

                    if (mSolver.getStatus() == CBaseMixedCriticalitySolver.Status.OPTIMAL || mSolver.getStatus() == CBaseMixedCriticalitySolver.Status.SUBOPTIMAL || mSolver.getStatus() == CBaseMixedCriticalitySolver.Status.HEURISTIC)
                    {
                        makespan += mInstance.getMakeSpan() ;
                        lateness += mInstance.getLateness() ;
                        cpu      += stopwatch.ElapsedMilliseconds ;
                    }
                    else
                    {
                        ++unsolved ;
                    }
                    
                    stopwatch.Reset() ;
                }

                mInstance.setMakeSpan(makespan / (mRepetitionCount - unsolved)) ;
                mInstance.setLateness(lateness / (mRepetitionCount - unsolved)) ;
                mInstance.setCPUTime((int)(cpu / (mRepetitionCount - unsolved))) ;
            }
            else
            {
                stopwatch.Start() ;

                mSolver.solve(mInstance) ;

                stopwatch.Stop() ;

                mInstance.setCPUTime((int)stopwatch.ElapsedMilliseconds) ;
            }

            outputSolution(mSolver.getStatus()) ;
        }
    }
}
