﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MCScheduling.Utils ;

namespace MCScheduling.MixedCriticality
{
    /**
     * Generates a random "worst-case" MC instance. A solution to each generated MC instance
     * corresponds to a (some) solution of the 3-Partition Problem.
     */
    class CWorstCaseMixedCriticalityInstanceGenerator
    {
        /**
         * 
         * @param instanceNamePrefix The prefix of the generated instances names.
         * @param instanceCount      The number of MC instances to generate.
         * @param numberOfClusters   The number of corresponding 3-partitions, i.e.
         *                           4-times the number of MC jobs will be generated
         *                           per instance.
         * @param clusterMakespan    The sum of each partition, i.e., the maximum
         *                           available execution time for the MC jobs
         *                           in each cluster.
         * 
         * @return The list of randomly generated MC instances.
         */
        public static List<CMixedCriticalityInstance> generate(string instanceNamePrefix, int instanceCount, int numberOfClusters, double clusterMakespan)
        {
            /* Each job in a partition must have its execution time at least equal to 1.0. */
            if (clusterMakespan < 3.0)
            {
                throw new ArgumentException("The partition sum has to be less or equal to 3.0!") ;
            }

            /** A pseudo-random number generator. */
            CUniformRandom randomizer = new CUniformRandom((ulong)DateTime.Now.Ticks) ;

            /* The list of generated instances. */
            List<CMixedCriticalityInstance> instanceList = new List<CMixedCriticalityInstance>(instanceCount) ;

            string   nameSuffixFormat = CDefaultMixedCriticalityInstanceGenerator.determineInstanceNameSuffix(instanceCount) ;
            double   releaseTime      = 0.0 ;
            double   deadline         = numberOfClusters * (clusterMakespan + 1) ;
            double[] executionTimes   = new double[3] ;

            /* Generates random mixed-criticality instances. */
            for (int iInstance = 1 ; iInstance <= instanceCount ; iInstance++)
            {
                /* A randomly generated instance. */
                CMixedCriticalityInstance instance = new CMixedCriticalityInstance(instanceNamePrefix + String.Format(nameSuffixFormat, iInstance)) ;

                double startTime = 0.0 ;

                /* Job generation loop. */
                for (int partition = 0 ; partition < numberOfClusters ; partition++)
                {
                    CMixedCriticalityJob job = new CMixedCriticalityJob("T" + (4 * partition + 1), releaseTime, deadline, 2, new double[] { 1.0, clusterMakespan + 1.0 }) ;
                    
                    job.setStartTime(startTime) ;
                    startTime = job.getCompletionTime(1) ;

                    instance.addJob(job) ;

                    executionTimes[0] = (int)(randomizer.NextDouble() * (clusterMakespan - 3.0) + 1.0) ;
                    executionTimes[1] = (int)(randomizer.NextDouble() * (clusterMakespan - executionTimes[0] - 2.0) + 1.0) ;
                    executionTimes[2] = (int)(clusterMakespan - executionTimes[0] - executionTimes[1]) ;
                    
                    for (int i = 0 ; i < 3 ; i++)
                    {
                        job = new CMixedCriticalityJob("T" + (4 * partition + i + 2), releaseTime, deadline, 1, new double[] { executionTimes[i] }) ;
                        
                        job.setStartTime(startTime) ;
                        startTime = job.getCompletionTime(1) ;

                        instance.addJob(job) ;
                    }
                }
                
                /* Validates generated instance. */
                if (!CDefaultMixedCriticalityInstanceGenerator.validateGeneratedInstance(instance))
                {
                    throw new InvalidOperationException("The instance generator does not work correctly!!!") ;
                }

                instance.setMakeSpan(0.0) ;
                instance.setLateness(0.0) ;

                instance.MakespanDataView = "0" ;
                instance.LatenessDataView = "0" ;

                /* Adds the job to the current instance. */
                instanceList.Add(instance) ;
            }

            return instanceList ;
        }
    }
}
