﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

/**
 * Contains the implementation of a mixed-criticality solver based on the dynamic programming
 * algorithm (similar to the one solving the traveling salesman problem).
 */
namespace MCScheduling.MixedCriticality.DP
{
    /**
     * An dynamic programming type algorithm for solving
     * mixed-criticality scheduling problems of minimal
     * size.
     */
    class CDPSolver : CBaseMixedCriticalitySolver
    {
        /* Infinity. */
        private static double INFINITY = double.MaxValue ;

        /* The C[i,S] is the shortest schedule starting with task [1] ending with task [i] that executes only tasks in S. */
        private CMixedCriticalityJob[,] C ;

        /* The P[i,S] is the preceeding task of task [i] in the shortest schedule from [1] to [i] that executes tasks in S. */
        private int[,] P ;

        /* The number of tasks. */
        private int N ;

        /* The minimum length of a found schedule. */
        private double minimumScheduleLength ;

        /* The resulting schedule: array of indices to jobs of the solved MC instance. */
        private int[] minimumSchedule ;

        /**
         * Solves the specified MC instance using dynamic programming, kind of algorithm
         * similar to TSP dynmaic programming solver.
         * 
         * @param instance The MC instance to solve.
         */
        public override void solve(CMixedCriticalityInstance instance)
        {
            CMixedCriticalityJob[] jobs = instance.getJobs() ;

            minimumScheduleLength = INFINITY ;

            minimumSchedule = new int[instance.getJobCount()] ;

            /* Resolves the instance each time with a different starting task. */
            for (int i = 0 ; i < jobs.Length ; i++)
            {
                CMixedCriticalityJob qq = jobs[0] ;
                jobs[0] = jobs[i] ;
                jobs[i] = qq ;

                if (solve(jobs, instance.getMaximumCriticality()))
                {
                    int j = 0, k = 0 ;

                    while (j < jobs.Length && minimumSchedule[j] != 0)
                    {
                        ++j ;
                    }

                    while (k < jobs.Length && minimumSchedule[k] != i)
                    {
                        ++k ;
                    }

                    minimumSchedule[j] = i ;
                    minimumSchedule[k] = 0 ;
                }

                qq      = jobs[0] ;
                jobs[0] = jobs[i] ;
                jobs[i] = qq ;
            }

            if (minimumScheduleLength < INFINITY)
            {
                reconstructSolution(minimumSchedule, instance) ;
            }
            else
            {
                setStatus(Status.INFEASIBLE) ;

                foreach (CMixedCriticalityJob job in instance)
                {
                    job.setStartTime(double.NaN) ;
                }
            }
        }

        /**
         * The inner method to solve the MC instance specified by the array of MC jobs
         * with the specified maximum criticality level.
         * 
         * @param jobs The MC jobs of the MC instance to solve.
         * @param maximumCriticality The criticality level of the specified MC instance.
         */
        private bool solve(CMixedCriticalityJob[] jobs, int maximumCriticality)
        {
            N = jobs.Length ;

            C = new CMixedCriticalityJob[N,(1 << N)] ;
            P = new int[N,(1 << N)] ;

            for(int k = 2 ; k <= N ; k++)
            {
                /* Creates a 1st dummy task that represents the 1st scheduled task of the MC instance. */
                CMixedCriticalityJob tmpJob  = new CMixedCriticalityJob("dummy", maximumCriticality, new double[maximumCriticality]) ;

                tmpJob.setStartTime(jobs[0].getReleaseTime()) ;
                tmpJob.setExecutionTimes(1, jobs[0].getCriticality(), jobs[0].getWCETs()) ;

                for (int i = jobs[0].getCriticality() + 1 ; i <= maximumCriticality ; i++)
                {
                    tmpJob.setExecutionTime(i, jobs[0].getLargestExecutionTime()) ;
                }
                
                /* The next task to be scheduled right after the 1st one. */
                CMixedCriticalityJob nextJob = jobs[k - 1] ;                

                nextJob.setStartTime(Math.Max(nextJob.getReleaseTime(), tmpJob.getCompletionTime(Math.Min(tmpJob.getCriticality(), nextJob.getCriticality())))) ;

                if (nextJob.getLargestCompletionTime() > nextJob.getDeadline())
                {
                    C[k - 1,(1 << (k - 1)) ^ 1] = null ;
                    P[k - 1,(1 << (k - 1)) ^ 1] = -1 ;
                }
                else
                {
                    for (int i = 1 ; i <= nextJob.getCriticality() ; i++)
                    {
                        tmpJob.setExecutionTime(i, nextJob.getCompletionTime(i)) ;
                    }

                    for (int i = 1 + nextJob.getCriticality() ; i <= maximumCriticality ; i++)
                    {
                        if (tmpJob.getExecutionTime(i - 1) > tmpJob.getExecutionTime(i))
                        {
                            tmpJob.setExecutionTime(i, tmpJob.getExecutionTime(i - 1)) ;
                        }
                        else
                        {
                            break ;
                        }
                    }

                    C[k - 1,(1 << (k - 1)) ^ 1] = tmpJob ;
                    P[k - 1,(1 << (k - 1)) ^ 1] = 0 ;
                }
            }

            int[] subset = new int[N] ;
            
            for(int s = 3 ; s <= N ; s++)
            {
                for(int i = 0 ; i < s - 1 ; i++)
                {
                    subset[i] = i ;
   	            }

                while(true)
                {          
                    int S = 1 ;

                    for(int i = 0 ; i < s - 1 ; i++)
                    {
                        S = S ^ (1 << (subset[i] + 1)) ;
                    }
	     
                    for(int i = 0 ; i < s - 1 ; i++)
                    {
                        int k = subset[i] + 2 ;

                        double prevScheduleLength = INFINITY ;

                        C[k - 1,S] = null ;
                        P[k - 1,S] = -1 ;

                        for(int j = 0 ; j < s - 1 ; j++)
                        {
                            int m = subset[j] + 2 ;

                            if(m != k)
                            {
                                if (C[m - 1,S ^ (1 << (k - 1))] != null)
                                {
                                    /* A dummy task that represents the shortest schedule of tasks in S ending with [m] task. */
                                    CMixedCriticalityJob lastJob  = (CMixedCriticalityJob)C[m - 1,S ^ (1 << (k - 1))].Clone() ;

                                    /* The next task [k] to be scheduled after the task [m]. */
                                    CMixedCriticalityJob nextJob = jobs[k - 1] ;

                                    nextJob.setStartTime(Math.Max(nextJob.getReleaseTime(), lastJob.getCompletionTime(Math.Min(lastJob.getCriticality(), nextJob.getCriticality())))) ;

                                    if (nextJob.getLargestCompletionTime() <= nextJob.getDeadline())
                                    {
                                        for (int criticality = 1 ; criticality <= nextJob.getCriticality() ; criticality++)
                                        {
                                            lastJob.setExecutionTime(criticality, nextJob.getCompletionTime(criticality)) ;
                                        }

                                        for (int criticality = 1 + nextJob.getCriticality() ; criticality <= maximumCriticality ; criticality++)
                                        {
                                            if (lastJob.getExecutionTime(criticality - 1) > lastJob.getExecutionTime(criticality))
                                            {
                                                lastJob.setExecutionTime(criticality, lastJob.getExecutionTime(criticality - 1)) ;
                                            }
                                            else
                                            {
                                                break ;
                                            }
                                        }

                                        if (lastJob.getLargestCompletionTime() < prevScheduleLength)
                                        {
                                            prevScheduleLength = lastJob.getLargestCompletionTime() ;

                                            C[k - 1,S] = lastJob ;
                                            P[k - 1,S] = m - 1 ;
                                        }
                                    }
                                }
                            }
                        }
                    }
    	     
                    /* Computes the next permutation of s bits in lexicographic order. */
                    int inext = (s - 1) - 1 ;

                    ++subset[inext] ;

                    while ((inext > 0) && (subset[inext] >= N - 1 - (s - 1) + 1 + inext))
                    {
                        ++subset[--inext] ;
                    }
 		            
                    /* If the last permutation has been processed, break out. */
                    if (subset[0] > (N - 1) - (s-1))
                    {
                        break ;
                    }

                    for (inext = inext + 1; inext < (s - 1); ++inext)
                    {
                        subset[inext] = subset[inext - 1] + 1 ;
                    }
                }
            }

            double optimum          = INFINITY ;
            int    lastScheduledJob = -1 ;

            for(int k = 2 ; k <= N ; k++)
            {
                if (C[k - 1,(1 << N) - 1] != null && C[k - 1,(1 << N) - 1].getLargestCompletionTime() < optimum)
                {
                    optimum = C[k - 1,(1 << N) - 1].getLargestCompletionTime() ;
                    lastScheduledJob = k - 1 ;
                }
            }

            if (optimum < minimumScheduleLength)
            {
                minimumScheduleLength = optimum ;

                int SS = (1 << N) - 1 ;
                int ii = jobs.Length ;

                minimumSchedule[--ii] = lastScheduledJob ;

                do
                {
                    int secondToLastScheduledJob = P[lastScheduledJob,SS] ;

                    SS ^= (1 << lastScheduledJob) ;

                    lastScheduledJob = secondToLastScheduledJob ;

                    minimumSchedule[--ii] = lastScheduledJob ;
                }
                while (lastScheduledJob  != 0) ;

                return true ;
            }

            return false ;
        }

        /**
         * Sets the parameters of the solution represented by the specified scheduledJobIndices up
         * in the specified mixed-criticality instance.
         * 
         * @param scheduledJobIndices The schedule representing a solution to the specified mixed-
         *                            criticality instance.
         * @param instance            The instance to be set up according to the solution represented 
         *                            by the specified array scheduledJobIndices.
         */
        private void reconstructSolution(int[] scheduledJobIndices, CMixedCriticalityInstance instance)
        {
            /* The lateness of the schedule represented by the evaluated chromosome. */
            double lateness = 0.0 ;

            /* The makespan of the schedule represented by the evaluated chromosome. */
            double makespan = 0.0 ;

            double[] completionTimeAtCriticality = new double[instance.getMaximumCriticality() + 1] ;

            CMixedCriticalityJob job = instance[scheduledJobIndices[0]] ;

            /* The first job starts at its release time. */
            job.setStartTime(job.getReleaseTime()) ;

            /* The completion time of the current 1 job schedule is actualized. */
            for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
            {
                completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
            }

            for (int i = 1 ; i < scheduledJobIndices.Length ; i++)
            {
                int maximumCompletionTimeCriticality = 1 ;

                /* The next scheduled job. */
                job = instance[scheduledJobIndices[i]] ;

                /* Determines the start time of the next job. */
                for (int criticality = 2 ; criticality <= job.getCriticality() ; criticality++)
                {
                    if (completionTimeAtCriticality[maximumCompletionTimeCriticality] < completionTimeAtCriticality[criticality] )
                    {
                        maximumCompletionTimeCriticality = criticality ;
                    }
                }

                job.setStartTime(Math.Max(completionTimeAtCriticality[maximumCompletionTimeCriticality], job.getReleaseTime())) ;

                /* Updates the completion times after the next job has been processed. */
                for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
                {
                    completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
                }

                /* Each delayed job increases the lateness. */
                lateness += Math.Max(0.0, job.getLargestCompletionTime() - job.getDeadline()) ;

                /* The makespan is also updated with each processed job. */
                if (makespan < job.getLargestCompletionTime())
                {
                    makespan = job.getLargestCompletionTime() ;
                }
            }

            /* Sets the makespan. */
            instance.setMakeSpan(makespan) ;

            /* Sets the lateness. */
            instance.setLateness(lateness) ;

            /* Feasible solution found. */
            if (lateness < 0.00001)
            {
                setStatus(Status.SUBOPTIMAL) ;
            }
            else
            {
                setStatus(Status.HEURISTIC) ;
            }
        }

        /**
         * Ignored. This solver is un-interruptable.
         */
        public override void interrupt()
        {
        }
    }
}
