﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using MCScheduling.GeneticAlgorithm ;
using MCScheduling.MixedCriticality ;
using MCScheduling.MixedCriticality.GA ;

/**
 * Contains the implementation of a mixed-criticality solver based on the genetic
 * algorithm.
 */
namespace MCScheduling.MixedCriticality.GA
{
    /**
     * The genetic algorihm based solver for the mixed-criticality scheduling.
     */
    public class CGeneticAlgorithmSolver : CBaseMixedCriticalitySolver
    {
        /** The configuration of the genetic algorithm used by this solver. */
        private CMixedCriticalityGeneticAlgorithm m_geneticAlgorithm ;

        /** If set to true the solver has been interrupted. */
        private bool m_interrupted ;

        /**
         * Creates a new genetic algorithm solver.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      solver.
         */
        public CGeneticAlgorithmSolver(CMixedCriticalityConfiguration configuration)
        {
            m_geneticAlgorithm = new CMixedCriticalityGeneticAlgorithm(configuration) ;
        }

        /**
         * Sets the specified configuration for the genetic algorithm
         * used by this solver.
         * 
         * @param configuration The configuration of the genetic 
         *                      algorithm.
         */
        public void setConfiguration(CMixedCriticalityConfiguration configuration)
        {
            m_geneticAlgorithm.setConfiguration(configuration) ;
        }

        /**
         * Returns the configuration of the genetic algorithm that
         * is used by this solver.
         * 
         * @return The configuration of the genetic algorithm that
         *         is used by this solver.
         */
        public CMixedCriticalityConfiguration getConfiguration()
        {
            return ((CMixedCriticalityConfiguration)m_geneticAlgorithm.getConfiguration()) ;
        }

        /**
         * Solves the specified mixed-criticality instance using genetic
         * algorithm.
         * 
         * @param instance The mixed-criticality instance to solve.
         */
        public override void solve(CMixedCriticalityInstance instance)
        {
            /* Sets the interruption flag to false. */
            m_interrupted = false ;

            getConfiguration().setInstanceToSolve(instance) ;
            getConfiguration().setGeneDistanceMeasurer(new CMixedCriticalityGeneticAlgorithm.CMixedCriticalityGeneDistanceMeasurer()) ;

            /* Initializes the algorithm. */
            m_geneticAlgorithm.reset() ;

            /* Time limit in miliseconds. */
            long timeLimit = (hasTimeLimit() ? getTimeLimit() * 1000 : long.MaxValue) ;
            
            /* Iteration limit. */
            long iterationLimit = (hasIterationLimit() ? getIterationLimit() : long.MaxValue) ;

            /* For elapsed time measurement. */
            System.Diagnostics.Stopwatch stopwatch = new System.Diagnostics.Stopwatch() ;

            stopwatch.Start() ;

            /* Let the evolution begin... */
            for (int iteration = 0 ; iteration < iterationLimit && !m_interrupted ; iteration++)
            {
                m_geneticAlgorithm.evolve() ;

                if (stopwatch.ElapsedMilliseconds > timeLimit)
                {
                    break ;
                }
            }

            stopwatch.Stop() ;

            reconstructSolution((CMixedCriticalityChromosome)m_geneticAlgorithm.getFittestChromosome(), instance) ;
        }

        /**
         * Sets the parameters of the solution represented by the specified chromosome up in the specified
         * mixed-criticality instance.
         * 
         * @param chromosome The chromosome representing a solution to the specified mixed-criticality instance.
         * @param instance   The instance to be set up according to the solution represented by the specified 
         *                   chromosome.
         */
        private void reconstructSolution(CMixedCriticalityChromosome chromosome, CMixedCriticalityInstance instance)
        {
            /* The schedule. */
            IGene[] genes = chromosome.getGenes() ;

            /* The lateness of the schedule represented by the evaluated chromosome. */
            double lateness = 0.0 ;

            /* The makespan of the schedule represented by the evaluated chromosome. */
            double makespan = 0.0 ;

            double[] completionTimeAtCriticality = new double[instance.getMaximumCriticality() + 1] ;

            CMixedCriticalityJob job = ((CMixedCriticalityGene)genes[0]).getScheduledJob() ;

            /* The first job starts at its release time. */
            job.setStartTime(job.getReleaseTime()) ;

            /* The completion time of the current 1 job schedule is actualized. */
            for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
            {
                completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
            }

            for (int geneIndex = 1 ; geneIndex < genes.Length ; geneIndex++)
            {
                int maximumCompletionTimeCriticality = 1 ;

                /* The next scheduled job. */
                job = ((CMixedCriticalityGene)genes[geneIndex]).getScheduledJob() ;

                /* Determines the start time of the next job. */
                for (int criticality = 2 ; criticality <= job.getCriticality() ; criticality++)
                {
                    if (completionTimeAtCriticality[maximumCompletionTimeCriticality] < completionTimeAtCriticality[criticality] )
                    {
                        maximumCompletionTimeCriticality = criticality ;
                    }
                }

                job.setStartTime(Math.Max(completionTimeAtCriticality[maximumCompletionTimeCriticality], job.getReleaseTime())) ;

                /* Updates the completion times after the next job has been processed. */
                for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
                {
                    completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
                }

                /* Each delayed job increases the lateness. */
                lateness += Math.Max(0.0, job.getLargestCompletionTime() - job.getDeadline()) ;

                /* The makespan is also updated with each processed job. */
                if (makespan < job.getLargestCompletionTime())
                {
                    makespan = job.getLargestCompletionTime() ;
                }
            }

            /* Sets the makespan. */
            instance.setMakeSpan(makespan) ;

            /* Sets the lateness. */
            instance.setLateness(lateness) ;

            /* Feasible solution found. */
            if (lateness < 0.00001)
            {
                setStatus(Status.SUBOPTIMAL) ;
            }
            else
            {
                setStatus(Status.HEURISTIC) ;
            }
        }

        /**
         * Interrupts the genetic algorithm solver.
         */
        public override void interrupt()
        {
            m_interrupted = true ;
        }
    }
}
