﻿using System;
using System.Collections ;
using System.Collections.Generic ;
using System.Linq;
using System.Text;
using MCScheduling.GeneticAlgorithm ;

namespace MCScheduling.MixedCriticality.GA
{
    /**
     * The CMixedCriticalityChromosome instance represents a schedule
     * for a mixed-criticality instance that is associated with its
     * genes which should be instances of CMixedCriticalityGene.
     * 
     * @author Tomas Tunys
     * 
     * @see CMixedCriticalityGene
     */
    public class CMixedCriticalityChromosome : IChromosome
    {
        /** The genes forming this chromosome. */
        private IGene[] m_genes ;

        /** The fitness value of the chromosome. */
        private double  m_fitnessValue ;

        /** 
         * The flag used to mark chromosome as being selected
         * for the mating pool.
         */
        private bool m_selected ;

        /**
         * Creates a new chromosome from the specified genes. Its
         * fitness value is initialy set to 0.
         * 
         * @param genes The gene values of the new chromosome.
         */
        public CMixedCriticalityChromosome(IGene[] genes) : this(genes, 0.0)
        {
        }

        /**
         * Creates a new chromosome from the specified genes. Its
         * fitness value is initialy set to the specified value.
         */
        private CMixedCriticalityChromosome(IGene[] genes, double fitnessValue)
        {
            m_genes = new IGene[genes.Length] ;

            setGenes(genes) ;
            setFitnessValue(fitnessValue) ;
            setSelectedForNextGeneration(false) ;
        }

        /**
         * Sets or gets the gene at the specified position.
         * 
         * @param index The zero-based position of the gene
         *              to be set or get.
         */
        public IGene this[int index]
        {
            get { return getGene(index) ; }
            set { setGene(value, index) ; }
        }

        /**
         * Sets the gene at the specified position.
         * 
         * @param gene The gene to be set.
         * @param index The zero-based position of 
         *              the gene to be set.
         */
        public void setGene(IGene gene, int index)
        {
            m_genes[index] = gene ;
        }

        /**
         * Returns the gene at the specified position.
         * 
         * @param index The zero-based position of 
         *              the gene.
         *              
         * @return The gene at the specified position.
         */
        public IGene getGene(int index)
        {
            return m_genes[index] ;
        }

        /**
         * Sets the genes of this chromosome to the specified
         * gene values.
         */
        public void setGenes(IGene[] genes)
        {
            Array.Copy(genes, m_genes, m_genes.Length) ;
        }

        /**
         * Returns the genes that form this chromosome.
         * 
         * @return The genes that form this chromosome.
         */
        public IGene[] getGenes()
        {
            return m_genes ;
        }

        /**
         * Returns the size of this chromosome, e.i. the 
         * number of genes it contains.
         * 
         * @return The size of this chromosome, e.i. the 
         * number of genes it contains.
         */
        public int getSize()
        {
            return m_genes.Length ;
        }

        /**
         * Sets the fitness score of this chromosome 
         * to the specified value.
         * 
         * @param value The chromosome's fitness score.
         */
        public void setFitnessValue(double value)
        {
            m_fitnessValue = value ;
        }

        /**
         * Returns the chromosome's fitness score.
         * 
         * @return The chromosome's fitness score.
         */
        public double getFitnessValue()
        {
            return m_fitnessValue ;
        }

        /**
         * Returns the copy of this chromosome.
         * 
         * @return The copy of this chromosome.
         */
        public object Clone()
        {
            return new CMixedCriticalityChromosome(m_genes, m_fitnessValue) ;
        }

        /**
         * Returns the gene enumerator of this chromosome.
         * 
         * @return The gene enumerator of this chromosome.
         */
        public IEnumerator<IGene> GetEnumerator()
        {
            return (IEnumerator<IGene>)m_genes.GetEnumerator() ;
        }

        /**
         * Returns the generic gene enumerator of this chromosome.
         * 
         * @return The gene enumerator of this chromosome.
         */
        IEnumerator IEnumerable.GetEnumerator()
        {
            return m_genes.GetEnumerator() ;
        }

        /**
         * Sets the selection flag of this chromosome to the specified
         * value.
         * 
         * @param selected If true the chromosome is marked as selected;
         *                 otherwise, it is unselected.
         */
        public void setSelectedForNextGeneration(bool selected)
        {
            m_selected = selected ;
        }

        /**
         * Returns true if this chromosome has been selected for the mating
         * pool; otherwise, it returns false.
         * 
         * @return True if this chromosome has been selected for the mating
         *         pool; otherwise, it returns false.
         */
        public bool isSelectedForNextGeneration()
        {
            return m_selected ;
        }

        /**
         * Releases any resources that this chromosome may use. This method is
         * invoked prior to its placement into a chromosome pool.
         */
        public void dispose()
        {
            setSelectedForNextGeneration(false) ;
        }       
    }
}