﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MCScheduling.GeneticAlgorithm ;

namespace MCScheduling.MixedCriticality.GA
{
    /**
     * The implementation of a genetic algorithm to solve mixed-criticality
     * scheduling problems.
     */
    public class CMixedCriticalityGeneticAlgorithm : CBaseGeneticAlgorithm
    {
        /**
         * Creates the genetic algorithm initialized with the specified
         * configuration that may be used to solve a mixed-criticality
         * scheduling problem.
         * 
         * @param configuration The configuration of the algorithm.
         */
        public CMixedCriticalityGeneticAlgorithm(CMixedCriticalityConfiguration configuration) : base(configuration) 
        {
        }

        #region >>> Initial Solution Generator. <<<

        #region >>> Priority Sorting Methods <<<

        /**
         * Comparer used to sort a list of job indices in earlist deadline first manner.
         * The ties are broken by the length of the execution time at common criticality,
         * the job with a longer common execution time is considered smaller.
         */
        private class CEarliestDeadlineFirst : System.Collections.Generic.IComparer<int>
        {
            /** 
             * The mixed-criticality instance such that the indices being sorted
             * correspond with the jobs in this instance.
             */
            CMixedCriticalityInstance m_instance ;

            /**
             * Creates a new CEarliestDeadlineFirst comparer.
             * 
             * @param instance The mixed-criticality instance associated with the index 
             *                 array being sorted using this comparer.
             */
            public CEarliestDeadlineFirst(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            /**
             * Compares two jobs with associated indexes in the respective mixed-criticality
             * instance in earliest deadline first manner.
             * 
             * @param i The index of the first job.
             * @param j The index of the second job.
             */
            public int Compare(int i, int j)
            {
                double diff = (m_instance[i].getDeadline() - m_instance[j].getDeadline()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }
                else
                {
                    int criticality = System.Math.Min(m_instance[i].getCriticality(), m_instance[j].getCriticality()) ;

                    diff = (m_instance[i].getExecutionTime(criticality) - m_instance[j].getExecutionTime(criticality)) ;

                    if (diff < 0)
                    {
                        return -1 ;
                    }
                    else if (diff > 0)
                    {
                        return 1 ;
                    }
                }

                return 0 ;
            }
        }

        /**
         * Comparer used to sort a list of job indices in earliest deadline last manner.
         * The ties are broken by the length of the execution time at common criticality,
         * the job with a smaller common execution time is considered smaller.
         */
        private class CEarliestDeadlineLast : System.Collections.Generic.IComparer<int>
        {
            /** 
             * The mixed-criticality instance such that the indices being sorted
             * correspond with the jobs in this instance.
             */
            CMixedCriticalityInstance m_instance ;

            /**
             * Creates a new CEarliestDeadlineLast comparer.
             * 
             * @param instance The mixed-criticality instance associated with the index 
             *                 array being sorted using this comparer.
             */
            public CEarliestDeadlineLast(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            /**
             * Compares two jobs with associated indexes in the respective mixed-criticality
             * instance in latest completion time last manner.
             * 
             * @param i The index of the first job.
             * @param j The index of the second job.
             */
            public int Compare(int i, int j)
            {
                double diff = (m_instance[j].getDeadline() - m_instance[i].getDeadline()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }
                else
                {
                    int criticality = System.Math.Min(m_instance[i].getCriticality(), m_instance[j].getCriticality()) ;

                    diff = (m_instance[j].getExecutionTime(criticality) - m_instance[i].getExecutionTime(criticality)) ;

                    if (diff < 0)
                    {
                        return -1 ;
                    }
                    else if (diff > 0)
                    {
                        return 1 ;
                    }
                    
                    return 0 ;
                }
            }
        }

        /**
         * Comparer used to sort a list of job indices in latest completion time first manner.
         * The ties are broken by the length of the execution time at common criticality,
         * the job with a longer common execution time is considered smaller.
         */
        private class CLatestCompletionTimeFirst : System.Collections.Generic.IComparer<int>
        {
            /** 
             * The mixed-criticality instance such that the indices being sorted
             * correspond with the jobs in this instance.
             */
            CMixedCriticalityInstance m_instance ;

            /**
             * Creates a new CLatestCompletionTimeFirst comparer.
             * 
             * @param instance The mixed-criticality instance associated with the index 
             *                 array being sorted using this comparer.
             */
            public CLatestCompletionTimeFirst(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            /**
             * Compares two jobs with associated indexes in the respective mixed-criticality
             * instance in latest completion time first manner.
             * 
             * @param i The index of the first job.
             * @param j The index of the second job.
             */
            public int Compare(int i, int j)
            {
                double diff = (m_instance[i].getDeadline() - m_instance[i].getLargestExecutionTime()) + (m_instance[j].getDeadline() - m_instance[j].getLargestExecutionTime()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }
                else
                {
                    int criticality = System.Math.Min(m_instance[i].getCriticality(), m_instance[j].getCriticality()) ;

                    diff = (m_instance[i].getExecutionTime(criticality) - m_instance[j].getExecutionTime(criticality)) ;

                    if (diff < 0)
                    {
                        return -1 ;
                    }
                    else if (diff > 0)
                    {
                        return 1 ;
                    }
                    
                    return 0 ;
                }
            }
        }

        /**
         * Comparer used to sort a list of job indices in latest completion time last manner.
         * The ties are broken by the length of the execution time at common criticality,
         * the job with a smaller common execution time is considered smaller.
         */
        private class CLatestCompletionTimeLast : System.Collections.Generic.IComparer<int>
        {
            /** 
             * The mixed-criticality instance such that the indices being sorted
             * correspond with the jobs in this instance.
             */
            CMixedCriticalityInstance m_instance ;

            /**
             * Creates a new CLatestCompletionTimeLast comparer.
             * 
             * @param instance The mixed-criticality instance associated with the index 
             *                 array being sorted using this comparer.
             */
            public CLatestCompletionTimeLast(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            /**
             * Compares two jobs with associated indexes in the respective mixed-criticality
             * instance in latest completion time last manner.
             * 
             * @param i The index of the first job.
             * @param j The index of the second job.
             */
            public int Compare(int i, int j)
            {
                double diff = (m_instance[j].getDeadline() - m_instance[j].getLargestExecutionTime()) + (m_instance[i].getDeadline() - m_instance[i].getLargestExecutionTime()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else
                {
                    int criticality = System.Math.Min(m_instance[i].getCriticality(), m_instance[j].getCriticality()) ;

                    diff = (m_instance[j].getExecutionTime(criticality) - m_instance[i].getExecutionTime(criticality)) ;

                    if (diff < 0)
                    {
                        return -1 ;
                    }
                    else if (diff > 0)
                    {
                        return 1 ;
                    }
                    
                    return 0 ;
                }
            }
        }

        #endregion >>> Priority Sorting Methods <<<

        /**
         * Generates the initial solution for the genetic algorithm. The initial 
         * solution is found by a heuristic priority-based algorithm using earliest
         * deadline first priority rule.
         * 
         * @param instance            The MC instance that is being solved by this
         *                            genetic algorithm.
         * @param jobPriorityOrdering The job priority-based ordering method.
         * 
         * @return The initial solution for the specified mixed-criticality instance.
         */
        protected CMixedCriticalityChromosome initialPriorityBasedSolution(CMixedCriticalityInstance instance, IComparer<int> jobPriorityOrdering)
        {
            int[]  initialSolution    = new int[instance.getJobCount()] ;
            int[]  jobInstanceIndices = new int[instance.getJobCount()] ;
            bool[] jobScheduled       = new bool[instance.getJobCount()] ;

            for (int i = 0 ; i < instance.getJobCount() ; i++)
            {
                jobInstanceIndices[i] = i ;
            }

            System.Array.Sort<int>(jobInstanceIndices, jobPriorityOrdering) ;

            double[] makespanAtCriticality = new double[instance.getMaximumCriticality() + 1] ;

            int numberOfScheduledJobs  = 0 ;
            int numberOfJobsToSchedule = instance.getJobCount() ;

            int scheduleTime = 0 ;

            while (numberOfScheduledJobs != numberOfJobsToSchedule)
            {
                int indexOfJobToSchedule = -1 ;

                while (++indexOfJobToSchedule < numberOfJobsToSchedule && (jobScheduled[indexOfJobToSchedule] || instance[jobInstanceIndices[indexOfJobToSchedule]].getReleaseTime() > scheduleTime)) ;

                if (indexOfJobToSchedule != numberOfJobsToSchedule)
                {
                    int maximumMakespanCriticality = 1 ;

                    for (int criticality = maximumMakespanCriticality + 1 ; criticality <= instance[jobInstanceIndices[indexOfJobToSchedule]].getCriticality() ; criticality++)
                    {
                        if (makespanAtCriticality[maximumMakespanCriticality] < makespanAtCriticality[criticality])
                        {
                            maximumMakespanCriticality = criticality ;
                        }
                    }

                    if (makespanAtCriticality[maximumMakespanCriticality] > scheduleTime)
                    {
                        scheduleTime = (int)makespanAtCriticality[maximumMakespanCriticality] ;
                    }

                    for (int criticality = 1 ; criticality <= instance[jobInstanceIndices[indexOfJobToSchedule]].getCriticality() ; criticality++)
                    {
                        makespanAtCriticality[criticality] = scheduleTime + instance[jobInstanceIndices[indexOfJobToSchedule]].getExecutionTime(criticality) ;
                    }

                    scheduleTime = (int)makespanAtCriticality[1] ;

                    jobScheduled[indexOfJobToSchedule] = true ;

                    initialSolution[numberOfScheduledJobs++] = jobInstanceIndices[indexOfJobToSchedule] ;
                }
                else
                {
                    ++scheduleTime ;
                }
            }

            IGene[] genes = new IGene[instance.getJobCount()] ;

            for (int i = 0 ; i < instance.getJobCount() ; i++)
            {
                genes[i] = new CMixedCriticalityGene(instance, initialSolution[i]) ;
            }

            return new CMixedCriticalityChromosome(genes) ;
        }

        /**
         * Creates an initial population for the start of the genetic algorithm. 
         * The initial population is generated this way: first 4 individuals
         * are generated according to 4 heuristics and the rest of the population
         * is generated randomly.
         * 
         * @param population The initial population to be filled up
         *                   with chromosomes.
         */
        public override void createInitialPopulation(CPopulation population)
        {
            if (!(getConfiguration() is CMixedCriticalityConfiguration))
            {
                throw new InvalidOperationException("The genetic algorithm does not have configuration of type CMixedCriticalityConfiguration!") ;
            }

            CMixedCriticalityConfiguration configuration = (CMixedCriticalityConfiguration)getConfiguration() ;

            CMixedCriticalityInstance instanceToSolve = configuration.getInstanceToSolve() ;

            population.addChromosome(initialPriorityBasedSolution(instanceToSolve, new CEarliestDeadlineFirst(instanceToSolve))) ;

            if (configuration.getPopulationSize() > 1)
            {
                population.addChromosome(initialPriorityBasedSolution(instanceToSolve, new CEarliestDeadlineLast(instanceToSolve))) ;
            }

            if (configuration.getPopulationSize() > 2)
            {
                population.addChromosome(initialPriorityBasedSolution(instanceToSolve, new CLatestCompletionTimeFirst(instanceToSolve))) ;
            }

            if (configuration.getPopulationSize() > 3)
            {
                population.addChromosome(initialPriorityBasedSolution(instanceToSolve, new CLatestCompletionTimeLast(instanceToSolve))) ;
            }

            IGene[] genes       = new CMixedCriticalityGene[configuration.getChromosomeSize()] ;
            int[]   permutation = new int[configuration.getChromosomeSize()] ;

            for (int i = 0 ; i < configuration.getChromosomeSize() ; i++)
            {
                genes[i] = new CMixedCriticalityGene(configuration.getInstanceToSolve(), i) ;
                permutation[i] = i ;
            }

            Random random = new Random() ;

            while (population.getSize() < configuration.getPopulationSize())
            {
                for (int i = permutation.Length - 1 ; i > 1 ; i--)
                {
                    int j = random.Next(i + 1) ;

                    int t = permutation[i] ;
                    permutation[i] = permutation[j] ;
                    permutation[j] = t ;

                    IGene g = genes[i] ;
                    genes[i] = genes[j] ;
                    genes[j] = g ;
                }

                population.addChromosome(new CMixedCriticalityChromosome(genes)) ;
            }
        }

        #endregion >>> Initial Solution Generator. <<<

        /**
         * The gene distance measurer for mixed-criticality genetic algorithm. It
         * may be used by some of its crossover/mutation operator.
         */
        public class CMixedCriticalityGeneDistanceMeasurer : IGeneDistanceMeasurer
        {
            /**
             * Returns the minimum time the 1st specified job delays the execution
             * of the 2nd specified job.
             * 
             * @param x The 1st mixed-criticality job.
             * @param y The 2nd mixed-criticality job.
             * 
             * @return The minimum time the 1st specified job delays the execution
             *         of the 2nd specified job.
             */
            public double distance(CMixedCriticalityJob x, CMixedCriticalityJob y)
            {
                return x.getExecutionTime(System.Math.Min(x.getCriticality(), y.getCriticality())) ;
            }

            /**
             * Returns the minimum time the job represented by the 1st specified 
             * gene delays the execution the job represented by the 2nd specified
             * gene.
             * 
             * @param x The 1st mixed-criticality gene.
             * @param y The 2nd mixed-criticality gene.
             * 
             * @return The minimum time the job represented by the 1st specified 
             *         gene delays the execution the job represented by the 2nd
             *         specified gene.
             */
            public double distance(IGene x, IGene y)
            {
                return distance(((CMixedCriticalityGene)x).getScheduledJob(), ((CMixedCriticalityGene)y).getScheduledJob()) ;
            }
        }
    }
}
