﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Gurobi ;

/**
 * Contains the implementation of a mixed-criticality solver based on the branch
 * and bound algorithm provided by Gurobi 4.5 library. To make it work a valid
 * licence needs to be acquired from http://www.gurobi.com/.
 */
namespace MCScheduling.MixedCriticality.MIP
{
    /**
     * The mixed integer programming solver for the mixed-criticality scheduling. 
     */
    class CMixedIntegerProgrammingSolver : CBaseMixedCriticalitySolver
    {
        /** If set to true the solver has been interrupted. */
        private bool m_interrupted ;

        /**
         * Callback for the MIP solver whose only purpose
         * is to interrupt the optimization process, if
         * needed.
         */
        private class CInterrupterCallback : GRBCallback
        {
            /** The owner of this callback. */
            private CMixedIntegerProgrammingSolver m_solver ;

            /**
             * Creates a new callback for the specified solver.
             * 
             * @param solver The MIP solver.
             */
            public CInterrupterCallback(CMixedIntegerProgrammingSolver solver)
            {
                m_solver = solver ;
            }

            /**
             * Invoked by Gurobi during the optimization process.
             * It just checks the interrupt flag, and if it is set
             * to true, it aborts the optimization.
             */
            protected override void Callback()
            {
                if (where == GRB.Callback.MIPNODE && m_solver.isInterrupted())
                {
                    Abort() ;
#if DEBUG
                    Console.WriteLine("[{0:hh:mm:ss}]: Optimization ABORTED!", System.DateTime.Now) ;
#endif
                }
            }
        }

        /**
         * Returns true if the solver has been
         * interrupted; otherwise, false.
         *
         * @return True if the solver has been
         *         interrupted.
         */
        public bool isInterrupted()
        {
            return m_interrupted ;
        }

        /**
         * Solves the specified mixed-criticality instance.
         * 
         * @param instance The mixed-criticality instance 
         *                 to be solved.
         */
        public override void solve(CMixedCriticalityInstance instance)
        {
            /* Sets the interruption flag to false. */
            m_interrupted = false ;

            try
	        {
		        GRBEnv environment = new GRBEnv() ;
#if !DEBUG
                /* Turns off the console output of the optimizer. */
		        environment.Set(GRB.IntParam.OutputFlag, 0) ;
#endif

                /* Sets up an iteration limit (if available). */
                if (hasIterationLimit())
                {
                    environment.Set(GRB.DoubleParam.IterationLimit, getIterationLimit()) ;
                }

                /* Sets up a time limit (if available). */
                if (hasTimeLimit())
                {
                    environment.Set(GRB.DoubleParam.TimeLimit, getTimeLimit()) ;
                }

                /* The lower bound for Cmax. */
                double LB = double.MinValue ;

                /* The upper bound for Cmax. */
		        double UB = double.MinValue ;

                foreach (CMixedCriticalityJob job in instance)
                {
                    if (LB < job.getReleaseTime())
                    {
                        LB = job.getReleaseTime() + job.getLargestExecutionTime() ;
                    }

                    if (UB < job.getDeadline())
                    {
                        UB = job.getDeadline() ;
                    }
                }
                    
                /* Commented upper bound approximations because they might not work. */
                /* UB = Math.Min(UB, approximateCmax(instance, new CLatestCompletionTimeFirst(instance))) ; */
                /* UB = Math.Min(UB, approximateCmax(instance, new CLatestCompletionTimeLast(instance))) ; */
                /* UB = Math.Min(UB, approximateCmax(instance, new CEarliestDeadlineLast(instance))) ; */   
                UB = Math.Min(UB, approximateCmax(instance, new CEarliestDeadlineFirst(instance))) ;                            
                
                /* A mixed-criticality MIP model. */
		        GRBModel model = new GRBModel(environment) ;

                /* The makespan of a schedule to be minimized. */
		        GRBVar Cmax = model.AddVar(LB, UB, 1.0, GRB.CONTINUOUS, "Cmax") ;
		
                /* Start times of the jobs in the specified mixed-criticality instance. */
		        GRBVar[]   s = new GRBVar[instance.getJobCount()] ;

                /* Decision variables: xij = 1 iff the i-th job precedes the j-th job. */
                GRBVar[][] x = new GRBVar[instance.getJobCount()][] ;

                /* Determines the upper bound and creates start time variables. */
		        for (int i = 0 ; i < instance.getJobCount() ; i++)
		        {			
			        s[i] = model.AddVar(instance[i].getReleaseTime(), instance[i].getDeadline() - instance[i].getLargestExecutionTime(), 0.0, GRB.CONTINUOUS, "s" + (i + 1).ToString()) ;
		        }

                /* Creates xij decision variables. */
		        for (int i = 0 ; i < instance.getJobCount() - 1 ; i++)
		        {
			        x[i] = new GRBVar[instance.getJobCount()]  ;

			        for (int j = i + 1 ; j < instance.getJobCount() ; j++)
			        {
				        x[i][j] = model.AddVar(0.0, 1.0, 0.0, GRB.BINARY, "x" + (i + 1).ToString() + (j + 1).ToString()) ;
			        }
		        }
	
                /* Updates the model to allow adding constraints. */
		        model.Update() ;

                /* Adds optimization constraints: s[i] + p[i][c[i]] <= Cmax. */
		        for (int i = 0 ; i < instance.getJobCount() ; i++)
		        {
			        model.AddConstr(s[i] + instance[i].getLargestExecutionTime() <= Cmax, "Optimality constraint for " + instance[i].getName()) ;
		        }

                /* Adds resource occupation constrains: x[i][j] = 1 --> s[i] + p[i][min(c[i],c[j])] <= s[j]
                 *                                      x[i][j] = 0 --> s[j] + p[i][min(c[i],c[j])] <= s[i]
                 */
                for (int i = 0 ; i < instance.getJobCount() - 1 ; i++)
		        {
			        for (int j = i + 1 ; j < instance.getJobCount() ; j++)
			        {
				        int criticality = Math.Min(instance[i].getCriticality(), instance[j].getCriticality()) ;
                        
				        model.AddConstr(s[i] + (-s[j]) + UB * x[i][j] <= UB - instance[i].getExecutionTime(criticality), instance[i].getName() + " precedes " + instance[j].getName()) ;
				        model.AddConstr(instance[j].getExecutionTime(criticality) <= s[i] + (-s[j]) + UB * x[i][j], instance[j].getName() + " precedes " + instance[i].getName()) ;
			        }
		        }

                /* Updates the model before optimization. */
		        model.Update() ;

                /* Sets the interrupt callback. */
                model.SetCallback(new CInterrupterCallback(this)) ;

                /* Optimizes the model. */
                model.Optimize() ;

                /* Gets the result of the optimization. */
		        int status = model.Get(GRB.IntAttr.Status) ;

                /* According to the status, solution is determined, and
                 * status variable set.
                 */
		        if (status == GRB.Status.OPTIMAL)
		        {
			        instance.setMakeSpan(model.Get(GRB.DoubleAttr.ObjVal)) ;

			        for (int i = 0 ; i < instance.getJobCount() ; i++)
			        {
				        instance[i].setStartTime(s[i].Get(GRB.DoubleAttr.X)) ;
			        }

                    setStatus(Status.OPTIMAL) ;
		        }
                /* The time limit has been exceeded. */
                else if (status == GRB.Status.TIME_LIMIT)
                {
                    try
                    {
                        instance.setMakeSpan(model.Get(GRB.DoubleAttr.ObjVal)) ;

                        for (int i = 0 ; i < instance.getJobCount() ; i++)
			            {
				            instance[i].setStartTime(s[i].Get(GRB.DoubleAttr.X)) ;
			            }

                        setStatus(Status.SUBOPTIMAL) ;
                    }
                    catch (GRBException)
                    {
                        setStatus(Status.TIME_LIMIT) ;
                    }
                }
                /* The iteration limit has been exceeded. */
                else if (status == GRB.Status.ITERATION_LIMIT)
                {
                    setStatus(Status.ITERATION_LIMIT) ;
                }
                /* The model has been proven to be infeasible. */
		        else if (status == GRB.Status.INFEASIBLE)
                {
			        setStatus(Status.INFEASIBLE) ;
                }
                /* An error has occured. */
                else
                {
                    setStatus(Status.ERROR) ;
                }
            }
	        catch(GRBException)
            {
                setStatus(Status.ERROR) ;
            }
        }

        /**
         * Interrupts the MIP solver.
         */
        public override void interrupt()
        {
            m_interrupted = true ;
        }

        private class CLatestCompletionTimeFirst : IComparer<int>
        {
            CMixedCriticalityInstance m_instance ;

            public CLatestCompletionTimeFirst(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            public int Compare(int i, int j)
            {
                double diff = (m_instance[i].getDeadline() - m_instance[i].getLargestExecutionTime()) + (m_instance[j].getDeadline() - m_instance[j].getLargestExecutionTime()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }

                return 0 ;
            }
        }

        private class CLatestCompletionTimeLast : IComparer<int>
        {
            CMixedCriticalityInstance m_instance ;

            public CLatestCompletionTimeLast(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            public int Compare(int i, int j)
            {
                double diff = (m_instance[j].getDeadline() - m_instance[j].getLargestExecutionTime()) + (m_instance[i].getDeadline() - m_instance[i].getLargestExecutionTime()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }

                return 0 ;
            }
        }

        private class CEarliestDeadlineFirst : IComparer<int>
        {
            CMixedCriticalityInstance m_instance ;

            public CEarliestDeadlineFirst(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            public int Compare(int i, int j)
            {
                double diff = (m_instance[i].getDeadline() - m_instance[j].getDeadline()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }
                else 
                {
                    int criticality = Math.Min(m_instance[i].getCriticality(), m_instance[j].getCriticality()) ;

                    diff = (m_instance[i].getExecutionTime(criticality) - m_instance[j].getExecutionTime(criticality)) ;

                    if (diff < 0)
                    {
                        return -1 ;
                    }
                    else if (diff > 0)
                    {
                        return 1 ;
                    }
                }

                return 0 ;
            }
        }

        private class CEarliestDeadlineLast : IComparer<int>
        {
            CMixedCriticalityInstance m_instance ;

            public CEarliestDeadlineLast(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            public int Compare(int i, int j)
            {
                double diff = (m_instance[j].getDeadline() - m_instance[i].getDeadline()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }

                return 0 ;
            }
        }

        private double approximateCmax(CMixedCriticalityInstance instance, IComparer<int> jobPriorityOrdering)
        {
            int[]  jobInstanceIndices = new int[instance.getJobCount()] ;
            bool[] jobScheduled       = new bool[instance.getJobCount()] ;

            for (int i = 0 ; i < instance.getJobCount() ; i++)
            {
                jobInstanceIndices[i] = i ;
            }

            Array.Sort<int>(jobInstanceIndices, jobPriorityOrdering) ;

            double[] makespanAtCriticality = new double[instance.getMaximumCriticality() + 1] ;

            int numberOfScheduledJobs  = 0 ;
            int numberOfJobsToSchedule = instance.getJobCount() ;

            int scheduleTime = 0 ;

            double maxCmax = double.MinValue ;

            while (numberOfScheduledJobs != numberOfJobsToSchedule)
            {
                int indexOfJobToSchedule = -1 ;

                while (++indexOfJobToSchedule < numberOfJobsToSchedule && (jobScheduled[indexOfJobToSchedule] || instance[jobInstanceIndices[indexOfJobToSchedule]].getReleaseTime() > scheduleTime)) ;

                if (indexOfJobToSchedule != numberOfJobsToSchedule)
                {
                    int maximumMakespanCriticality = 1 ;

                    for (int criticality = maximumMakespanCriticality + 1 ; criticality <= instance[jobInstanceIndices[indexOfJobToSchedule]].getCriticality() ; criticality++)
                    {
                        if (makespanAtCriticality[maximumMakespanCriticality] < makespanAtCriticality[criticality])
                        {
                            maximumMakespanCriticality = criticality ;
                        }
                    }

                    if (makespanAtCriticality[maximumMakespanCriticality] > scheduleTime)
                    {
                        scheduleTime = (int)makespanAtCriticality[maximumMakespanCriticality] ;
                    }

                    for (int criticality = 1 ; criticality <= instance[jobInstanceIndices[indexOfJobToSchedule]].getCriticality() ; criticality++)
                    {
                        makespanAtCriticality[criticality] = scheduleTime + instance[jobInstanceIndices[indexOfJobToSchedule]].getExecutionTime(criticality) ;

                        if (maxCmax < makespanAtCriticality[criticality])
                        {
                            maxCmax = makespanAtCriticality[criticality] ;
                        }
                    }

                    scheduleTime = (int)makespanAtCriticality[1] ;

                    jobScheduled[indexOfJobToSchedule] = true ;

                    ++numberOfScheduledJobs ;
                }
                else
                {
                    ++scheduleTime ;
                }
            }

            return maxCmax ;
        }
    }
}