﻿using MCScheduling.SimulatedAnnealing ;
using MCScheduling.GeneticAlgorithm ;

namespace MCScheduling.MixedCriticality.SA
{
    /**
     * The simulated annealing solver for the mixed-criticality scheduling. 
     */
    public class CMixedCriticalitySimulatedAnnealing : CBaseSimulatedAnnealing<CSimulatedAnnealingState>
    {
        private CMixedCriticalityInstance m_instance ;

        private CSimulatedAnnealingConfiguration m_configuration ;

        /**
         * Sets up the configuration of this simulated annealing algorithm.
         * 
         * @param configuration The settings of this simulated annealing algorithm.
         */
        public void setConfiguration(CSimulatedAnnealingConfiguration configuration)
        {
            m_configuration = configuration ;
        }

        /**
         * Returns the configuration of simulated annealing algorithm.
         * 
         * @return The configuration of simulated annealing algorithm.
         */
        public CSimulatedAnnealingConfiguration getConfiguration()
        {
            return m_configuration ;
        }

        #region >>> Simulated Annealing Components. <<<

        /**
         * Classic exponential distribution acceptor.
         */
        public class CExponentialAcceptor : IAcceptor
        {
            public double Probability(double distance)
            {
                if (distance <= 0.0)
                {
                    return 1 ;
                }
                else
                {
                    return System.Math.Exp(-distance) ;
                }
            }
        }

        /**
         * Lower energy - lower makespan / lateness - better solution.
         */
        public class CLowerBetterComparer : IComparer
        {
            /**
             * Compares the two specified evaluations (energies)
             * of simulated annealing states.
             * 
             * @param x The first state's energy.
             * @param y The second state's energy.
             * 
             * @return -1 if x is greater then y, 0 if they are
             *         equal, 1 otherwise.
             */
            public int Compare(double x, double y)
            {
                return (int)(y - x) ;
            }
        }

        /**
         * The proportional cooler works the way that the temperature
         * at the next time instant is proportional to the temperature
         * in the previous. The factor is the parameter of the cooler.
         */
        public class CProportionalCooler : ICooler
        {
            /** The factor by which is the temperature decreased each iteration. */
            private double m_factor ;

            /** The factor for the current iteration. */
            private double m_currentFactor ;

            /**
             * Creates a new CProportionalCooler instance with the specified
             * factor of the temperature decrease.
             */
            public CProportionalCooler(double factor)
            {
                m_factor = factor ;
                m_currentFactor = 1.0 ;
            }

            /**
             * Invoked when the simulated annealing optimization
             * is about to be executed.
             */
            public void Reset()
            {
                m_currentFactor = m_factor ;
            }

            /**
             * Invoked in each iteration of the simulated annealing
             * algorithm.
             * 
             * @param temperature The current temperature of the simulated
             *                    annealing process.
             *
             * @return The temperature for the next iteration of the simulated
             *         annealing process.
             */
            public double Cooldown(double temperature)
            {
                m_currentFactor *= m_factor ;

                return (temperature * m_currentFactor) ;
            }
        }

        /**
         * Evaluates the energy of the given state as a convex sum of the makespan
         * and lateness of the corresponding solution schedule.
         */
        public class CMinimumMakespanLatenessEvaluator : IEvaluator<CSimulatedAnnealingState>
        {
            /** The associated mixed-criticality instance of the states being evaluated. */
            private CMixedCriticalityInstance m_instance ;

            /** Used to build up a schedule. */
            private double[] m_completionTimeAtCriticality ;

            /** The lateness factor. */
            private double m_latenessFactor ;

            /** The makespan factor. */
            private double m_makespanFactor ;

            /**
             * Creates the MinimumMakespanLatenessEvaluator with makespan
             * and lateness factors set to 0.5.
             * 
             * @param instance The mixed-criticality instance associated
             *                 with a solution.
             */
            public CMinimumMakespanLatenessEvaluator(CMixedCriticalityInstance instance) : this(instance, 0.5, 0.5)
            {
            }

            /**
             * Creates the MinimumMakespanLatenessEvaluator for the specified 
             * lateness, makespan weighting factor.
             * 
             * @param instance       The mixed-criticality instance associated
             *                       with a solution.
             * @param makespanFactor The weight of makespan.
             * @param latenessFactor The weight of lateness.
             */
            public CMinimumMakespanLatenessEvaluator(CMixedCriticalityInstance instance, double makespanFactor, double latenessFactor)
            {   
                setInstance(instance) ;
                setMakespanFactor(makespanFactor) ;
                setLatenessFactor(latenessFactor) ;

                m_completionTimeAtCriticality = new double[100] ;
            }

            /**
             * Returns the makespan factor.
             * 
             * @return The makespan factor.
             */
            public double getMakespanFactor()
            {
                return m_makespanFactor ;
            }

            /**
             * Sets the makespan factor.
             * 
             * @param makespanFactor The new makespan factor value.
             */
            public void setMakespanFactor(double makespanFactor)
            {
                m_makespanFactor = makespanFactor ;
            }

            /**
             * Returns the lateness factor.
             * 
             * @return The lateness factor.
             */
            public double getLatenessFactor()
            {
                return m_latenessFactor ;
            }

            /**
             * Sets the lateness factor.
             * 
             * @param latenessFactor The new lateness factor value.
             */
            public void setLatenessFactor(double latenessFactor)
            {
                m_latenessFactor = latenessFactor ;
            }

            /**
             * Returns the mixed-criticality instance associated
             * with the solution.
             */
            public CMixedCriticalityInstance getInstace()
            {
                return m_instance ;
            }

            /**
             * Sets the mixed-criticality instance associated
             * with the solution.
             * 
             * @param instance The MC instance associated with the solution.
             */
            public void setInstance(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            /**
             * Calculates and returns the value of the energy (quality) of 
             * the specified state.
             * 
             * @param state The state to be evaluated.
             * 
             * @return The value of the energy (quality) of the state. 
             */
            public double Evaluate(CSimulatedAnnealingState state)
            {
                /* The lateness of the schedule represented by the state. */
                double lateness = 0.0 ;

                /* The makespan of the schedule represented by the state. */
                double makespan = 0.0 ;

                for (int i = 1 ; i < m_instance.getMaximumCriticality() ; i++)
                {
                    m_completionTimeAtCriticality[i] = 0.0 ;
                }

                CMixedCriticalityJob job = m_instance[state.ScheduledJobIndex[0]] ;

                /* The first job starts at its release time. */
                job.setStartTime(job.getReleaseTime()) ;

                state.StartTime[0] = (int)(job.getStartTime()) ;
                state.Lateness[0]  = (int)(job.getLargestCompletionTime() - job.getDeadline()) ;

                /* The completion time of the current 1 job schedule is actualized. */
                for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
                {
                    m_completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
                }

                for (int i = 1 ; i < state.Length ; i++)
                {
                    int maximumCompletionTimeCriticality = 1 ;

                    /* The next scheduled job. */
                    job = m_instance[state.ScheduledJobIndex[i]] ;

                    /* Determines the start time of the next job. */
                    for (int criticality = 2 ; criticality <= job.getCriticality() ; criticality++)
                    {
                        if (m_completionTimeAtCriticality[maximumCompletionTimeCriticality] < m_completionTimeAtCriticality[criticality] )
                        {
                            maximumCompletionTimeCriticality = criticality ;
                        }
                    }

                    job.setStartTime(System.Math.Max(m_completionTimeAtCriticality[maximumCompletionTimeCriticality], job.getReleaseTime())) ;

                    state.StartTime[i] = (int)(job.getStartTime()) ;
                    state.Lateness[i]  = (int)(job.getLargestCompletionTime() - job.getDeadline()) ;

                    /* Updates the completion times after the next job has been processed. */
                    for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
                    {
                        m_completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
                    }

                    /* Each delayed job increases the lateness. */
                    lateness += System.Math.Max(0.0, job.getLargestCompletionTime() - job.getDeadline()) ;

                    /* The makespan is also updated with each processed job. */
                    if (makespan < job.getLargestCompletionTime())
                    {
                        makespan = job.getLargestCompletionTime() ;
                    }
                }

                return (m_latenessFactor * lateness + m_makespanFactor * makespan) ;  
            }
        }

        /**
         * The perturber used to change one state into one of its "neighbouring" state. Basically
         * it works such that it randomly selects a tardy job and tries to schedule it earlier
         * to fulfil its deadline.
         */
        public class CPerturber : IPerturber<CSimulatedAnnealingState>
        {
            /* The mixed-criticality instance associated with states being perturbed. */
            CMixedCriticalityInstance m_instance ;

            /* The pseudo-random number generator. */
            IRandomizer m_randomizer ;

            /* The priority queue to find out latest jobs fast. */
            CPriorityQueue<int> m_latestJobs ;

            /* The comparer used with the priority queue. */
            LatenessComparer m_latenessComparer ;

            /* The list of highest lateness values. */
            long[] m_highestLatenessValues ;

            private class LatenessComparer : System.Collections.Generic.IComparer<int>
            {
                /* the latenesses of the jobs being compared. */
                int[] m_latenesses ;

                /**
                 * Sets the latenesses of the jobs to be compared.
                 * 
                 * @param latenesses The lateness of the jobs being
                 *                   compared.
                 */
                public void setLatenesses(int[] latenesses)
                {
                    m_latenesses = latenesses ;
                }

                public int Compare(int x, int y)
                {
                    return m_latenesses[x].CompareTo(m_latenesses[y]) ;
                }
            }

            /**
             * The mixed-criticality instance associated with the perturbed
             * state.
             * 
             * @param instance   The mixed-criticality instance associated
             *                   with states being perturbed.
             * @param randomizer The random number generator.
             */
            public CPerturber(CMixedCriticalityInstance instance, IRandomizer randomizer)
            {
                m_instance              = instance ;
                m_randomizer            = randomizer ;
                m_latenessComparer      = new LatenessComparer() ;
                m_latestJobs            = new CPriorityQueue<int>(m_latenessComparer, System.Math.Max(1, instance.getJobCount() / 4), true) ;
                m_highestLatenessValues = new long[System.Math.Max(1, instance.getJobCount() / 4)] ;
                //m_latestJobs            = new CPriorityQueue<int>(m_latenessComparer, 1, true) ;
                //m_highestLatenessValues = new long[1] ;
            }

            /**
             * Moves randomly from the specified state to a neighbouring
             * state.
             * 
             * @param state The current state to move from.
             */
            public void perturb(CSimulatedAnnealingState state)
            {
                m_latenessComparer.setLatenesses(state.Lateness) ;

                for (int i = 0 ; i < state.Length ; i++)
                {
                    m_latestJobs.push(i) ;
                }

                /* The jobs with the highest lateness values. */
                System.Collections.Generic.IList<int> latestJobIndices = m_latestJobs.elements() ;

                long minimumLateness = long.MaxValue ;

                /* Finds out the lowest between the highest lateness values. */
                for (int i = 0 ; i < latestJobIndices.Count ; i++)
                {
                    m_highestLatenessValues[i] = (long)state.Lateness[latestJobIndices[i]] ;

                    if (minimumLateness > m_highestLatenessValues[i])
                    {
                        minimumLateness = m_highestLatenessValues[i] ;
                    }                    
                }

                /* There might be no late job. */
                if (minimumLateness < 0)
                {
                    minimumLateness = -minimumLateness ;
                }

                m_highestLatenessValues[0] += minimumLateness ;

                /* Makes the latenesses positive and calculates the cumulative lateness. */
                for (int i = 1 ; i < latestJobIndices.Count ; i++)
                {
                    m_highestLatenessValues[i] += m_highestLatenessValues[i - 1] + minimumLateness ;
                }

                int selectedLatestJobIndex = System.Array.BinarySearch<long>(m_highestLatenessValues, (long)(m_randomizer.NextDouble() * m_highestLatenessValues[latestJobIndices.Count - 1])) ;

                /* Happens most of the time - a negative index is the lowest index of an individual with higher cumulative fitness 
                 * then the randomly generated value - with an appropriate conversion, of course. */
                if (selectedLatestJobIndex < 0)
                {
                    selectedLatestJobIndex = -(selectedLatestJobIndex + 1) ;
                }

                int jobToRescheduleIndex = latestJobIndices[selectedLatestJobIndex] ;

                /* The release time of the job to be rescheduled. */
                int releaseTime = (int)m_instance[state.ScheduledJobIndex[jobToRescheduleIndex]].getReleaseTime() ;
                int startTime   = state.StartTime[jobToRescheduleIndex] ;

                int lowestSchedulableIndex = System.Math.Max(0, jobToRescheduleIndex - 1) ;

                /* Finds out the smallest possible position for the job to be rescheduled at. */
                while (lowestSchedulableIndex > 0 && state.StartTime[lowestSchedulableIndex] >= releaseTime)
                {
                    --lowestSchedulableIndex ;
                }

                int jobsNewPosition = m_randomizer.Next(lowestSchedulableIndex, jobToRescheduleIndex) ;

                int tmp = state.ScheduledJobIndex[jobToRescheduleIndex] ;

                while (jobToRescheduleIndex != jobsNewPosition)
                {
                    state.ScheduledJobIndex[jobToRescheduleIndex] = state.ScheduledJobIndex[jobToRescheduleIndex - 1] ;

                    --jobToRescheduleIndex ;
                }

                state.ScheduledJobIndex[jobsNewPosition] = tmp ;

                /* !!! The start time and lateness values are recounted later when the state is evaluated. !!! */
                /* !!! The only important piece of information is the schedule represented by the state.ScheduledJobIndex array. !!! */
            }
        }

        #endregion >>> Simulated Annealing Components. <<<

        #region >>> Initial Solution Generator. <<<

        /**
         * Comparer used to sort a list of job indices in earlist deadline first manner.
         * The ties are broken by the length of the execution time at common criticality,
         * the job with a longer common execution time is considered smaller.
         */
        private class CEarliestDeadlineFirst : System.Collections.Generic.IComparer<int>
        {
            /** 
             * The mixed-criticality instance such that the indices being sorted
             * correspond with the jobs in this instance.
             */
            CMixedCriticalityInstance m_instance ;

            /**
             * Creates a new CEarliestDeadlineFirst comparer.
             * 
             * @param instance The mixed-criticality instance associated with the index 
             *                 array being sorted using this comparer.
             */
            public CEarliestDeadlineFirst(CMixedCriticalityInstance instance)
            {
                m_instance = instance ;
            }

            /**
             * Compares two jobs with associated indexes in the respective mixed-criticality
             * instance in earliest deadline first manner.
             * 
             * @param i The index of the first job.
             * @param j The index of the second job.
             */
            public int Compare(int i, int j)
            {
                double diff = (m_instance[i].getDeadline() - m_instance[j].getDeadline()) ;

                if (diff < 0)
                {
                    return -1 ;
                }
                else if (diff > 0)
                {
                    return 1 ;
                }
                else
                {
                    int criticality = System.Math.Min(m_instance[i].getCriticality(), m_instance[j].getCriticality()) ;

                    diff = (m_instance[i].getExecutionTime(criticality) - m_instance[j].getExecutionTime(criticality)) ;

                    if (diff < 0)
                    {
                        return -1 ;
                    }
                    else if (diff > 0)
                    {
                        return 1 ;
                    }
                }

                return 0 ;
            }
        }

        /**
         * Generates the initial solution for the simulated annealing algorithm.
         * The initial solution is found by a heuristic priority-based algorithm
         * using earliest deadline first priority rule.
         * 
         * @return The initial state of the simulated annealing.
         */
        protected override CSimulatedAnnealingState InitialSolution()
        {
            int[]  ScheduledJobIndex  = new int[m_instance.getJobCount()] ;
            int[]  JobStartTimes      = new int[m_instance.getJobCount()] ;
            int[]  JobLateness        = new int[m_instance.getJobCount()] ;
            int[]  jobInstanceIndices = new int[m_instance.getJobCount()] ;
            bool[] jobScheduled       = new bool[m_instance.getJobCount()] ;

            for (int i = 0 ; i < m_instance.getJobCount() ; i++)
            {
                jobInstanceIndices[i] = i ;
            }

            System.Array.Sort<int>(jobInstanceIndices, new CEarliestDeadlineFirst(m_instance)) ;

            double[] makespanAtCriticality = new double[m_instance.getMaximumCriticality() + 1] ;

            int numberOfScheduledJobs  = 0 ;
            int numberOfJobsToSchedule = m_instance.getJobCount() ;

            int scheduleTime = 0 ;

            while (numberOfScheduledJobs != numberOfJobsToSchedule)
            {
                int indexOfJobToSchedule = -1 ;

                while (++indexOfJobToSchedule < numberOfJobsToSchedule && (jobScheduled[indexOfJobToSchedule] || m_instance[jobInstanceIndices[indexOfJobToSchedule]].getReleaseTime() > scheduleTime)) ;

                if (indexOfJobToSchedule != numberOfJobsToSchedule)
                {
                    int maximumMakespanCriticality = 1 ;

                    for (int criticality = maximumMakespanCriticality + 1 ; criticality <= m_instance[jobInstanceIndices[indexOfJobToSchedule]].getCriticality() ; criticality++)
                    {
                        if (makespanAtCriticality[maximumMakespanCriticality] < makespanAtCriticality[criticality])
                        {
                            maximumMakespanCriticality = criticality ;
                        }
                    }

                    if (makespanAtCriticality[maximumMakespanCriticality] > scheduleTime)
                    {
                        scheduleTime = (int)makespanAtCriticality[maximumMakespanCriticality] ;
                    }

                    JobStartTimes[numberOfScheduledJobs] = (int)scheduleTime ;
                    JobLateness[numberOfScheduledJobs]   = (int)(scheduleTime + m_instance[jobInstanceIndices[indexOfJobToSchedule]].getLargestExecutionTime() - m_instance[jobInstanceIndices[indexOfJobToSchedule]].getDeadline()) ;

                    for (int criticality = 1 ; criticality <= m_instance[jobInstanceIndices[indexOfJobToSchedule]].getCriticality() ; criticality++)
                    {
                        makespanAtCriticality[criticality] = scheduleTime + m_instance[jobInstanceIndices[indexOfJobToSchedule]].getExecutionTime(criticality) ;
                    }

                    scheduleTime = (int)makespanAtCriticality[1] ;

                    jobScheduled[indexOfJobToSchedule] = true ;

                    ScheduledJobIndex[numberOfScheduledJobs++] = jobInstanceIndices[indexOfJobToSchedule] ;
                }
                else
                {
                    ++scheduleTime ;
                }
            }

            return new CSimulatedAnnealingState(ScheduledJobIndex, JobStartTimes, JobLateness) ;
        }

        #endregion >>> Initial Solution Generator. <<<

        /**
         * Transformes the specified state to a state from its nearest neighbourhood.
         * 
         * @param state The state to move.
         * @param temperature The current value of the temperature parameter.
         */
        protected override void moveToNeighbour(ref CState state, double temperature)
        {
            for (int i = 0 ; i < state.State.Length ; i++)
            {
                base.moveToNeighbour(ref state, temperature) ;
            }
        }

        /**
         * Sets the MC instance to be solved by running a simulated annealing.
         *
         * @param instance The MC instance to solve.
         */
        public void setInstanceToSolve(CMixedCriticalityInstance instance)
        {
            m_instance = instance ;

            setAcceptor(new CExponentialAcceptor()) ;
            setComparer(new CLowerBetterComparer()) ;
            setCooler(new CProportionalCooler(getConfiguration().getCoolingFactor())) ;
            setEvaluator(new CMinimumMakespanLatenessEvaluator(m_instance, getConfiguration().getMakespanFactor(), getConfiguration().getLatenessFactor())) ;
            setPerturber(new CPerturber(m_instance, getRandomGenerator())) ;
            setMinimumTemperature(getConfiguration().getMinimumTemperature()) ;
            setMaximumTemperature(getConfiguration().getMaximumTemperature()) ;
        }
    }
}
