﻿using MCScheduling.GeneticAlgorithm ;

/**
 * Contains the implementation of a (generic) simulated annealing
 * algorithm that may be used to solve different kind of optimization
 * problems.
 */
namespace MCScheduling.SimulatedAnnealing
{
    /**
     * The base simulated annealing algorithm.
     * 
     * @param T The type of solution.
     * 
     * @author Tomas Tunys
     */
    public abstract class CBaseSimulatedAnnealing<T> where T : ICloneable<T>
    {
        /** The energy evaluator of states. */
        private IEvaluator<T> m_evaluator ;

        /** The comparer for state energies. */
        private IComparer m_comparer ;

        /** The acceptance distribution of states according to energy. */
        private IAcceptor m_acceptor ;

        /** The cooling function of temperature. */
        private ICooler m_cooler ;

        /** Neigbouring state generator. */
        private IPerturber<T> m_perturber ;

        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /** Interruption flag. */
        private bool m_interrupted ;

        /** The maximum (starting) temperature. */
        private double m_maximumTemperature ;

        /** The minimum (ending) temperature. */
        private double m_minimumTemperature ;

        /** The current neigbouring state. */
        private CState m_neighbour ;

        /** The currently best state. */
        private CState m_bestState ;

        #region >>> Setters and Getters. <<<

        /**
         * Sets the evaluator.
         * 
         * @param evaluator The evaluator of states.
         */
        public void setEvaluator(IEvaluator<T> evaluator)
        {
            m_evaluator = evaluator ;
        }

        /**
         * Sets the fitness score comparer.
         * 
         * @param comparer The comparer.
         */
        public void setComparer(IComparer comparer)
        {
            m_comparer = comparer ;
        }

        /**
         * Sets the acceptor probability distribution method.
         * 
         * @param acceptor The acceptor.
         **/
        public void setAcceptor(IAcceptor acceptor)
        {
            m_acceptor = acceptor ;
        }

        /**
         * Sets the temperature cooling method.
         * 
         * @param cooler The temperature cooling method.
         */
        public void setCooler(ICooler cooler)
        {
            m_cooler = cooler ;
        }

        /**
         * Sets the neighbour generating method.
         *
         * @param perturber A method generating a neighbouring
         *                  state given a current state.
         */
        public void setPerturber(IPerturber<T> perturber)
        {
            m_perturber = perturber ;
        }

        /**
         * Sets the maximum (starting) temperature for the simulated
         * annealing.
         * 
         * @param maximumTemperature The maximum temperature value.
         */
        public void setMaximumTemperature(double maximumTemperature)
        {
            m_maximumTemperature = maximumTemperature ;
        }

        /**
         * Returns the maximum (starting) temperature of the simulated
         * annealing algorithm.
         * 
         * @return The maximum (starting) temperature of the simulated
         * annealing algorithm.
         */
        public double getMaximumTemperature()
        {
            return m_maximumTemperature ;
        }

        /**
         * Sets the minimum (ending) temperature for the simulated
         * annealing.
         * 
         * @param minimumTemperature The minimum temperature value.
         */
        public void setMinimumTemperature(double minimumTemperature)
        {
            m_minimumTemperature = minimumTemperature ;
        }

        /**
         * Returns the minimum (ending) temperature of the simulated
         * annealing algorithm.
         * 
         * @return The minimum (ending) temperature of the simulated
         * annealing algorithm.
         */
        public double getMinimumTemperature()
        {
            return m_minimumTemperature ;
        } 

        /**
         * Sets the pseudo-random number generator for the simulated
         * annealing algorithm.
         * 
         * @param randomizer The random number generator.
         */
        public void setRandomGenerator(IRandomizer randomizer)
        {
            m_randomizer = randomizer ;
        }

        /**
         * Returns the random number generator used by the simulated
         * annealing algorithm.
         * 
         * @return The random number generator.
         */
        public IRandomizer getRandomGenerator()
        {
            return m_randomizer ;
        }

        /**
         * Sets the interrupted flag on the specified value. If the value
         * is true, and the optimization process is currently running, then
         * it is interrupted.
         * 
         * @param interrupted If true, then the running optimization process
         *                    is interrupted.
         */
        public void setInterrupted(bool interrupted)
        {
            m_interrupted = interrupted ;
        }

        /**
         * Returns the state of the interrupted flag.
         * 
         * @return True if the simulated annealing has been interrupted; otherwise
         *              it returns false.
         */
        public bool isInterrupted()
        {
            return m_interrupted ;
        }

        #endregion >>> Setters and Getters. <<<

        /**
         * The state keeps information about
         * the energy of the actual state it
         * wraps.
         */
        protected class CState
        {
            /** The state. */
            public T State ;

            /** The state's fitness score. */
            public double Energy ;
        }

        /**
         * Initializes the basic simulated annealing method.
         */
        protected CBaseSimulatedAnnealing() : this(new CMersenneTwister((ulong)System.DateTime.Now.Ticks))
        {
        }

        /**
         * Initializes the base simulated annealing method.
         * 
         * @param randomizer The pseudo-random number generator to use.
         */
        protected CBaseSimulatedAnnealing(IRandomizer randomizer)
        {
            if (randomizer == null)
            {
                throw new System.ArgumentException("The random number generator cannot be null!") ;
            }
            else 
            {
                m_randomizer = randomizer ;
                m_bestState  = new CState() ;
                m_neighbour  = new CState() ;
            }
        }
        
        /**
         * The metropolis algorithm. Returns a neighbouring state of
         * the specified state.
         * 
         * @param state The current state.
         * @param temperature The temperature influencing probability
         *                    of acceptance of a neighbouring state.
         */
        protected virtual void moveToNeighbour(ref CState state, double temperature)
        {
            /* First makes a copy of the current solution... */
            m_neighbour.State.Clone(state.State) ;

            /* ... to be randomly tweaked a instant later... */
            m_perturber.perturb(m_neighbour.State) ;

            /* ... and finally evaluated. */
            m_neighbour.Energy = m_evaluator.Evaluate(m_neighbour.State) ;
                
            /* The neighbour is accepted with the probability which depends on the difference between
             * the solutions and the current temperature.
             */
            if (m_randomizer.NextDouble() < m_acceptor.Probability((m_neighbour.Energy - state.Energy) / temperature))
            {
                CState swappee = m_neighbour ; m_neighbour = state ; state = swappee ;
            }
        }       

        /**
         * Solves a problem using simulated annealing. The process is
         * stopped after the specified number of iterations has been
         * exceded or the specified number of miliseconds have passed,
         * the temperature has dropped down to minimum temperature, or
         * the process has been interrupted.
         * 
         * @param iterationLimit The maximum number of iterations.
         * @param timeLimit      The maximum number of miliseconds.
         * 
         * @return The state that represents a solution, or the last
         *         state when the algorithm has been stopped.
         */
        public T solve(long iterationLimit, long timeLimit)
        {
            double temperature = m_maximumTemperature ;

            m_interrupted = false ;

            CState currentState = new CState() ;

            currentState.State = InitialSolution() ;
            currentState.Energy = m_evaluator.Evaluate(currentState.State) ;

            m_bestState.State = (T)currentState.State.Clone() ;
            m_bestState.Energy = currentState.Energy ;

            m_neighbour.State = (T)currentState.State.Clone() ;
            m_neighbour.Energy = currentState.Energy ;

            /* For elapsed time measurement. */
            System.Diagnostics.Stopwatch stopwatch = new System.Diagnostics.Stopwatch() ;

            stopwatch.Start() ;

            while (iterationLimit-- > 0 && timeLimit > stopwatch.ElapsedMilliseconds && temperature > m_minimumTemperature && !m_interrupted)
            {
                moveToNeighbour(ref currentState, temperature) ;

                if (m_comparer.Compare(m_bestState.Energy, currentState.Energy) < 0)
                {
                    m_bestState.State.Clone(currentState.State) ;
                    m_bestState.Energy = currentState.Energy ;
                }

                temperature = m_cooler.Cooldown(temperature) ;
            }

            stopwatch.Stop() ;

            return m_bestState.State ;
        }

        /**
         * Returns the initial solution.
         * 
         * @return The initial solution.
         */
        protected abstract T InitialSolution() ;
    }
}
