﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.Utils
{
    /**
     * A exponential distribution random number generator. The method of
     * generation random number is based on inverse transform sampling:
     * -ln(U)/lambda, where U is drawn from the uniform distribution 
     * on the unit interval (0,1), and lambda is the parameter of
     * the distribution, so-called rate parameter.
     */
    class CExponentialRandom
    {
        /** The parameter of the exponential distribution. */
        private double m_lambda ;

        /** The uniform random number generator. */
        private CUniformRandom m_uniform ;

        /** The rate parameter setter and getter. */
        public double Lambda 
        {
            set { m_lambda = value ; }
            get { return m_lambda ; }
        }

        /** 
         * Creates a random number generator with exponential distribution
         * with lambda set to 1.0 (default).
         */
        public CExponentialRandom() : this(1.0)
        {
        }

        /** 
         * Creates a random number generator with exponential distribution
         * with lambda set to the specified value.
         * 
         * @param lambda The distribution parameter.
         */
        public CExponentialRandom(double lambda)
        {
            m_lambda  = 1.0 ;
            m_uniform = new CUniformRandom() ; 
        }

        /** 
         * Creates a random number generator with exponential distribution
         * with lambda set to 1.0 (default) and initializes the generator
         * with the specified seed.
         * 
         * @param seed The value used to initialize the generator.
         */
        public CExponentialRandom(ulong seed) : this(1.0, seed)
        {
        }

        /** 
         * Creates a random number generator with exponential distribution
         * with lambda set to the specified value, and initializes the generator
         * with the specified seed.
         * 
         * @param lambda The distribution parameter.
         * @param seed   The value used to initialize the generator.
         */
        public CExponentialRandom(double lambda, ulong seed)
        {
            m_lambda  = lambda ;
            m_uniform = new CUniformRandom(seed) ; 
        }

        /**
         * Generates a random non-negative decimal drawn from exponential
         * distribution.
         * 
         * @return A random non-negative decimal drawn from exponential
         *         distribution
         */
        public double Next()
        {
            return -Math.Log(1.0 - m_uniform.NextDouble()) / m_lambda ;
        }
    }
}
